#audio_draw_window.py    12Dec2022  crs, from audio_window.py
#                        08Nov2022  crs, Author
"""
Provide simple drawing graphical window with audio feedback
to facilitate use and examination by the visually impaired.
Uses turtle to facilitate cursor movement within screen
Adapted from audio_window to concentrate on figure drawing
as well as presentation.
"""
import os
import tkinter as tk

from select_trace import SlTrace
from speaker_control import SpeakerControlLocal
from wx_grid_path import GridPath
from braille_cell import BrailleCell
from magnify_info import MagnifyInfo, MagnifyDisplayRegion
from adw_front_end import AdwFrontEnd

class AudioDrawWindow:



    
    def __init__(self,
        title=None, speaker_control=None,
        win_width=800, win_height=800,
        grid_width=40, grid_height=25,
        x_min=None, y_min=None,
        line_width=1, color="black",
        pos_check_interval= .1,
        pos_rep_interval = .1,
        pos_rep_queue_max = 1,
        visible = True,
        enable_mouse = False,
        pgmExit=None,
        blank_char=",",
        drawing=False,
        show_marked=False,
        shift_to_edge=True,
        silent=False,
        look_dist=2,
        menu_str="",
        key_str="",
        mag_info=None,
        iy0_is_top=True,        # OBSOLETE
                 ):
        """ Setup audio window
        :speaker_control: (SpeakerControlLocal) local access to centralized speech making
        :win_width: display window width in pixels
            default: 800
        :win_height: display window height in pixels
            default: 800
        :x_min: minimum coordinate
                default: 0
        :y_min: minimum coordinate
                default: 0
        :grid_width: braille width in cells
            default: 40
        :grid_height: braille width in cells
            default: 25
        :title: window title
        :pos_check_interval: time between checks/reporting on
                cursor position  default: .1 seconds
        :pos_rep_interval: minimum time between reports
                default: .5 seconds
        :pos_rep_force_output: force output
                default: Force
        :pos_rep_queue_max: maximum position report queue maximum
                default: 4
        :visible: cells are visible
                default: True - visible
        :blank_char: replacement for non-trailing blanks
                    default "," to provide a "mostly blank",
                    non-compressed blank character for the
                    braille graphics
                    default: "," dot 2.
        :shift_to_edge: shift picture to edge/top
                    to aid in finding
                    default: True - shift
        :drawing: We're in drawing mode
                 default: False - no drawing
        :show_marked: Make marked cells
                shown even when invisible
                default: True
        :silent: Suppress talking / audio
                default: False
        :menu_str:  Initial menu commands
                    default=None
                    Note that interpretation is case insensitive
                    Format: <Menu Letter><option letters>; as separator
                            e.g. n:sh;d:d  for Navigator silent, help
        :enable_mouse: mouse cursor operation enabled
                    default: False                                        Drawing: start drawing
        :key_str:  Initial key symbol command string
                    Each command separated by ";" to facilitate
                    recognition of multi charater symbols
                    such as "up", "down" Note that
                    interpretation is case insensitive
        :look_dist: Max number of cells to look ahead and report
                    default: 1
        :iy0_is_top: OBSOLETE
        """
        
        # direction for digit pad
        if title is None:
            title = "AudioDrawWindow"
        self.title = title
        control_prefix = "AudioDraw"
        self.win_width = win_width
        self.win_height = win_height
        self.grid_width = grid_width
        self.grid_height = grid_height
        self.cell_height = win_height/self.grid_height
        self.pgmExit = pgmExit
        if x_min is None:
            x_min = 0
        if y_min is None:
            y_min = 0
        # create the feedback window
        mw = tk.Tk()
        self.title = title
        mw.title(title)
        self.mw = mw
        if speaker_control is None:
            SlTrace.lg("Creating own SpeakerControl")
            speaker_control = SpeakerControlLocal(win=mw)
        self.speaker_control = speaker_control
        #mw.withdraw()


        
        win_print_frame = tk.Frame(mw)
        win_print_frame.pack(side=tk.TOP)
        win_print_entry = tk.Entry(
                                win_print_frame, width=50)
        win_print_entry.pack(side=tk.TOP)
        self.win_print_entry = win_print_entry
        
        canvas = tk.Canvas(mw, width=self.win_width, height=self.win_height)
        canvas.pack()
        self.canvas = canvas
        self.update()        # Force display
        SlTrace.lg(f"canvas width: {canvas.winfo_width()}")
        SlTrace.lg(f"canvas height: {canvas.winfo_height()}")
        self._visible = visible
        self.fte = AdwFrontEnd(self, title=title, silent=silent, color=color)
        self.set_x_min(x_min)
        self.set_y_min(y_min)
        self.set_x_max(x_min + win_width)
        self.set_y_max(y_min + win_height)
        self.set_drawing(drawing)
        self.speak_text(title)

        self.escape_pressed = False # True -> interrupt/flush
        self.cells = {}         # Dictionary of cells by (ix,iy)
        self.set_cell_lims()
        self.do_talking = True      # Enable talking
        self.logging_speech = True  # Output speech to log/screen
        if mag_info is None:
            top_region = MagnifyDisplayRegion(
                x_min=self.get_x_min(), y_min=self.get_y_min(),
                x_max=self.get_x_max(),
                y_max=self.get_y_max(),
                nrows=self.grid_height,
                ncols=self.grid_width,
                
                )
            mag_info = MagnifyInfo(top_region=top_region)
        else:
            if mag_info.description:
                title = mag_info.description
                mw.title(title)
                self.title = title
                self.speak_text(f"Magnification of {title}")
        mag_info.display_window = self    # Magnification setup
        self.mag_info = mag_info
        SlTrace.lg(f"\nAudioDrawWindow() mag_info:\n{mag_info}\n")
        self.mag_selection_tag = None       # mag selection canvas tag, if one
        self.is_selected = False            # Flag as not yet selected
        self.pos_history = []       # position history (ix,iy)
        self.pos_rep_force_output = False
        self.goto_travel_list = []    # Memory of where we have gone
        self.goto_travel_list_index = None
        self.cell_history = []          # History of all movement
        self.set_grid_path()
        self.running = True         # Set False to stop
        self.mw.focus_force()
        self.blank_char = blank_char
        self.shift_to_edge = shift_to_edge
        self.set_look_dist(look_dist)
        self.fte.do_complete(menu_str=menu_str, key_str=key_str)
        #self.pos_check()            # Startup possition check loop
        self.update()     # Make visible


    def exit(self, rc=None):
        """ Main exit if creating magnifications
        """
        if rc is None:
            rc = 0
        if self.pgmExit is not None:
            self.pgmExit()      # Use supplied pgmExit
            
        SlTrace.lg("AudoDrawWindow.exit")
        SlTrace.onexit()    # Force logging quit
        os._exit(0)

    def set_look_dist(self, look_dist):
        self.look_dist = look_dist

    def get_look_dist(self):
        return self.look_dist

    def silence(self):
        """ Check if silent mode
        """
        return self.fte.silence()

    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9            
        """
        self.win_print(msg)
        if self.is_silent():
            if dup_stdout:
                SlTrace.lg(msg)
            return
        
        self.speaker_control.speak_text(msg=msg, msg_type=msg_type,
                             dup_stdout=dup_stdout,
                             rate=rate, volume=volume)

    def speak_text_stop(self):
        """ Stop ongoing speach, flushing queue
        """
        self.speaker_control.speak_text_stop()
        
    def trav_len(self, ix, iy, dir_x, dir_y, require_cell=True):
        """ Find the travel length
        This is the number of steps from cell(ix,iy) in 
        direction (dir_x,dir_y) while still in self.cells[]
        traversal is stopped at traversed cells - in
        self.goto_cell_list
        :ix: cell x-index
        :iy: cell y_index
        :dir_x: x change each step
        :dir_y: y change each step
        :require_cell: Require cell in traversal default: True - required
        :returns: steps_forward, steps_backward, -1,-1
                 if cell, itself is not in figure(self.cells)
        """
        if require_cell and (ix,iy) not in self.cells:
            return -1,-1       # Not in figure
        
        if dir_x == 0 and dir_y == 0:
            return 0,0
        
        tlen_forward = 0
        ix_f = ix
        iy_f = iy
        while True:
            ix_f += dir_x
            iy_f += dir_y
            if (ix_f,iy_f) not in self.cells:
                break
            if (ix_f,iy_f) in self.get_goto_cell_list():
                break
            tlen_forward += 1   # traversal extended

        # Get backend of line
        tlen_backward = 0
        ix_b = ix
        iy_b = iy
        
        while True:
            ix_b -= dir_x
            iy_b -= dir_y
            if (ix_b,iy_b) not in self.cells:
                break
            if (ix_b,iy_b) in self.get_goto_cell_list():
                break
            tlen_backward += 1   # traversal extended
        return tlen_forward,tlen_backward
             
    def draw_cells(self, cells=None, show_points=False):
        """ Display braille cells on canvas
        :cells: list or dictionary cells to draw
        :show_points: instead of braille, show sample points
        """
        SlTrace.lg("draw_cells")
        SlTrace.lg(f"x_min:{self.get_x_min()} y_min: {self.get_y_min()}")
        if cells is None:
            cells = self.get_cells()
        else:
            if not isinstance(cells, dict):
                cs = {}
                for cell in cells:
                    cs[(cell.ix,cell.iy)] = cell
                cells = cs
            self.cells = cells      # Copy
        min_x, max_y, max_x,min_y = self.bounding_box()
        if min_x is not None:            
            SlTrace.lg(f"Lower left: min_x:{min_x} min_y:{min_y}")
            SlTrace.lg(f"Upper Right: max_x:{max_x} max_y:{max_y}")
        SlTrace.lg(f"{len(cells)} cells")
        for cell in cells.values():
            self.display_cell(cell)
            cell.mtype = cell.MARK_UNMARKED
        self.update()
        if min_x is not None:
            self.set_xy((min_x,min_y))
            x,y = self.get_xy()
            #self.pos_check(x=x,  y=y)
        self.set_grid_path()
        self.pos_history = []       # Clear history
        self.set_scanning()         # Setup positioning info
        self.update()
                    
    def print_braille(self, title=None, shift_to_edge=None):
        """ Output braille display
        :title: title default: self.title
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        """
        if title is None:
            title = self.title
        if title is not None:
            print(title)
        if shift_to_edge is None:
            shift_to_edge = self.shift_to_edge
        if shift_to_edge:
            self.find_edges()
            left_edge = self.left_edge
            right_edge = self.right_edge
            top_edge = self.top_edge
            bottom_edge = self.bottom_edge
        else:
            left_edge = 0
            top_edge = self.grid_height-1

        braille_text = ""
        for iy in range(top_edge, bottom_edge):
            line = ""
            for ix in range(left_edge, right_edge):
                cell_ixy = (ix,iy)
                if cell_ixy in self.cells:
                    cell = self.cells[cell_ixy]
                    color = cell.color_string()
                    line += color[0]
                else:
                    line += " "
            line = line.rstrip()
            if self.blank_char != " ":
                line = line.replace(" ", self.blank_char)
            ###print(f"{iy:2}", end=":")
            braille_text += line + "\n"
        SlTrace.lg(braille_text)
        self.mw.clipboard_clear()
        if title is not None:
            self.mw.clipboard_append(f"\n{title}\n")
        self.mw.clipboard_append(braille_text)

    def print_cells(self, title=None):
        """ Display current braille in a window
        """
        if title is not None:
            print(title)
        for ix in range(self.grid_width):
            for iy in range(self.grid_height):
                cell_ixy = (ix,iy)
                if cell_ixy in self.cells:
                    cell = self.cells[cell_ixy]
                    SlTrace.lg(f"ix:{ix} iy:{iy} {cell_ixy}"
                               f" {cell._color}"
                          f" rect: {self.get_cell_rect_tur(ix,iy)}"
                          f"  win rect: {self.get_cell_rect_win(ix,iy)}")
        SlTrace.lg("")

    def find_edges(self):
        """Find  top, left, bottom, right non-blank edges
        so we can shift picture to left,top for easier
        recognition
        :returns: left_edge, top_edge, right_edge, bottom_edge
                    Also sets self.left_edge,...
        """
        left_edge,top_edge, right_edge,bottom_edge = self.bounding_box_ci()
        if bottom_edge is None:
            bottom_edge = self.grid_height-1
                    
        if left_edge > 0:           # Give some space
            left_edge -= 1
        if left_edge > 0:
            left_edge -= 1
        self.left_edge = left_edge
        
        if top_edge > 0:           # Give some space
            top_edge -= 1
        if bottom_edge  < self.grid_height-1:
            bottom_edge += 1        
        self.top_edge = top_edge
        self.bottom_edge = bottom_edge
        
        self.left_edge = left_edge
        self.right_edge = right_edge
         
        return left_edge, top_edge, right_edge, bottom_edge

        
    def set_cursor_pos_win(self, x=0, y=0, quiet=False):
        """ Set mouse cursor position in win(canvas) coordinates
        :x: x-coordinate (win(canvas))
        :y: y-coordinate (win(canvas)) 
        :quiet: Don't announce legal moves
                default:False
        """
        """ Set mouse cursor position in canvas coordinates
        :x: x-coordinate (canvas win)
        :y: y-coordinate (canvas win)
        :quiet: Don't announce legal moves
                default: False
        """
        if not self.running:
            return
        
        self.set_xy((x,y))      # All we do is set coordinate memory
        loc_ixiy = self.get_ixy_at()
        self.pos_history.append(loc_ixiy)   # location history
        SlTrace.lg(f"pos_history:{loc_ixiy}", "pos_tracking")
        cell_ixiy = self.get_cell_at()
        if cell_ixiy is not None:
            self.cell_history.append(cell_ixiy)
            if not self.is_drawing():   # If we're not drawing
                self.mark_cell(cell_ixiy)   # Mark cell if one
        self.cursor_update()

        if not self.mw.winfo_exists():
            return 
        
        self.update()
        if not quiet:
            self.pos_check(force_output=True)

    def move_to_ixy(self, ix=None, iy=None):
        """ Move to grid (cell) ix,iy
        :ix: ix index default: current ix
        :iy: iy index default: current iy
        """
        ix_cur,iy_cur = self.get_ixy_at()
        if ix is None:
            ix = ix_cur 
        if iy is None:
            iy = iy_cur
        win_xc,win_yc = self.get_cell_center_win(ix,iy)
        self.move_to(win_xc,win_yc)



    def move_to(self, x,y, quiet=False):
        """ Move to window loc
        Stop at edges, with message
        :x:  win x-coordinate
        :y: win y-coordinate
        :quiet: Don't announce legal move
                default:False
        """
        self.set_cursor_pos_win(x=x, y=y, quiet=quiet)

    def is_inbounds_ixy(self, *ixy):
        """ Check if ixy pair is in bounds
        :ixy: if tuple ix,iy pair default: current location
              else ix,iy indexes
            ix: cell x index default current location
            iy: cell y index default current location
        :returns: True iff in bounds else False
        """
        ix_cur,iy_cur = self.get_ixy_at()
        if len(ixy) ==  0:
            ix,iy = ix_cur,iy_cur
        elif len(ixy) == 1 and isinstance(ixy[0], tuple):
            ix,iy = ixy[0]
        elif len(ixy) == 2:
            ix,iy = ixy
        else:
            raise Exception(f"bad is_inbounds_ixy args: {ixy}")
        if ix is None:
            ix = ix_cur
        if iy is None:
            iy = iy_cur
        if ix < 0 or ix >= len(self.cell_xs):
            return False
         
        if iy < 0 or iy >= len(self.cell_ys):
            return False
        
        return True 
                       
    def set_cell_lims(self):
        """ create cell boundary values bottom through top
         so:
             cell_xs[0] == left edge
             cell_xs[grid_width] == right edge
             cell_ys[0] == top edge
             cell_ys[grid_height] == bottom edge
        """
         
        self.cell_xs = []
        self.cell_ys = []

        for i in range(self.grid_width+1):
            x = int(self.get_x_min() + i*self.win_width/self.grid_width)
            self.cell_xs.append(x)
        for i in reversed(range(self.grid_height+1)):
            y = int(self.get_y_min() + i*self.win_height/self.grid_height)
            self.cell_ys.append(y)
            
    def get_ix_min(self):
        """ get minimum ix on grid
        :returns: min ix
        """
        return 0

    def get_ix_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.grid_width-1

    def get_iy_min(self):
        """ get minimum iy on grid
        :returns: min iy
        """
        return 0

    def get_iy_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.grid_height-1

    def get_mag_info(self):
        """ Get magnification info storage
        """
        return self.mag_info

    def get_speaker_control(self):
        """ Get speech control
        """
        return self.speaker_control

        

    def bounding_box(self):
        """ canvas coordinates which bound displayed figure
        :returns: min_x, max_y, max_x, min_y  (upper left) (lower right)
                    None,None,None,None if no figure
        """
        min_ix, max_iy, max_ix, min_iy = self.bounding_box_ci()
        if min_ix is None:
            return None,None,None,None      # No figure
        
        min_x, max_y, _, _ = self.get_cell_rect_tur(min_ix,max_iy)
        _, _, max_x, min_y = self.get_cell_rect_tur(max_ix,min_iy)
        return min_x,max_y, max_x, min_y
    
    def bounding_box_ci(self, cells=None, add_edge=None):
        """ cell indexes which bound the list of cells
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :returns: 
                    None,None,None,None if no figure
                    upper left ix,iy  lower right ix,iy
        """
        if cells is None:
            if not hasattr(self, "cells"):
                return None,None,None,None         # Not yet setup
            cells = list(self.cells.keys())
        
        ix_min, iy_max, ix_max,iy_min = None,None,None,None
        for cell in cells:
            cell_ixy = cell if isinstance(cell, tuple) else (cell.ix,cell.iy)
            cell_ix, cell_iy = cell_ixy
            if ix_min is None or cell_ix < ix_min:
                ix_min = cell_ix
            if ix_max is None or cell_ix > ix_max:
                ix_max = cell_ix
            if iy_min is None or cell_iy < iy_min:
                iy_min = cell_iy
            if iy_max is None or cell_iy > iy_max:
                iy_max = cell_iy

        if ix_min is None:
            ix_min = 0
        if ix_max is None:
            ix_max = self.grid_width-1
        if iy_min is None:
            iy_min = 0
        if iy_max is None:
            iy_max = self.grid_height-1
        
        if add_edge is not None:    # Extend/Shrink box
            ext_ix_min = ix_min - add_edge
            bd_ix_max = self.get_ix_max() # protect against too large of negative add_edge
            if ext_ix_min > bd_ix_max: ext_ix_min = bd_ix_max
            ix_min = max(ext_ix_min, self.get_ix_min()) # limit to bounds

            ext_iy_min = iy_min - add_edge
            bd_iy_max = self.get_iy_max()
            if ext_iy_min > bd_iy_max: ext_iy_min = bd_iy_max
            iy_min = max(ext_iy_min, self.get_iy_min())
            
            ext_ix_max = ix_max + add_edge
            bd_ix_min = self.get_ix_min()
            if ext_ix_max < bd_ix_min: ext_ix_max = bd_ix_min
            ix_max = min(ext_ix_max, self.get_ix_max())
            
            ext_iy_max = iy_max + add_edge
            bd_iy_min = self.get_iy_min()
            if ext_iy_max < bd_iy_min: ext_ix_max = bd_iy_min
            iy_max = min(ext_iy_max, self.get_iy_max())
            
        return ix_min,iy_min, ix_max,iy_max
            


    def erase_cell(self, cell):
        """ Erase cell
        :cell: BrailleCell
        """
        if cell is None:
            return
        
        # Remove current items, if any
        if cell.canv_items:
            for item_id in cell.canv_items:
                self.canvas.delete(item_id)
            self.update()       # Make change visible
        cell.canv_items = []

    def erase_pos_history(self):
        """ Remove history, undo history marking
        """
        canvas = self.canvas
        for cell_ixy in self.pos_history:
            if not cell_ixy in self.cells:
                continue
            cell = self.cells[cell_ixy]
            self.display_cell(cell)
            '''
            for item_id in cell.canv_items:
                item_type = canvas.type(item_id)
                if item_type == "rectangle":
                    canvas.itemconfigure(item_id, fill='')
                    canvas.tag_lower(item_id)
            '''
        self.pos_history = []
        canvas = self.canvas
        if self.mag_selection_tag is not None:
            canvas.itemconfig(self.mag_selection_tag, outline="red")
        self.is_selected = False         # Flag as unselected
            
        self.update()
                
    def display_reposition_hack(self, cx1,cx2,cy1,cy2, force=False):
        """ ###TFD HACK to reposition cell dispaly
            move x1,x2,y1,y2
        """
        return cx1,cx2,cy1,cy2

    def annotate_cell(self, cell_ixy=None, color=None,
                      outline="blue", outline_width=2,
                      text=None):
        """ Annotate cell to highlight it
        Possibly for perimeter viewing
        :cell_xy: ix,iy tuple default: current location
        :color: rectangle color default: no fill
        :outline: add outline color
                    default: no special outline
        :outline_width: outline width
                    default: 2
        :text: added text default: no text added
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        cell = self.get_cell_at_ixy(cell_ixy=cell_ixy)
        if cell is None:
            return     # Just ignore if missing
        
        self.deannotate_cell(cell_ixy=cell_ixy) #???
        self.display_cell(cell=cell)
        canvas = self.canvas
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas(cell_ixy)
        if outline is not None:
            canv_item = canvas.create_rectangle(cx1,cy1,cx2,cy2,
                                fill=color,
                                 outline=outline,
                                 width=outline_width)
            cell.canv_items.append(canv_item)
            
        self.update()    # So we can see it now 
            
        return 

    def deannotate_cell(self, cell_ixy=None, color=None,
                      outline=None, text=None):
        """ deAnnotate cell, restoring cell to normal
        :cell_xy: ix,iy tuple default: current location
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        cell = self.get_cell_at_ixy(cell_ixy=cell_ixy)
        if cell is not None:
            self.display_cell(cell=cell)
            
        return      # Just ignore if missing
    
    def display_cell(self, cell, show_points=False):
        """ Display cell
        :cell: BrailleCell
        :show_points: show points instead of braille
                default: False --> show braille dots
        """
        self.erase_cell(cell)
        if not cell.is_visible():
            return              # Nothing to show
        
        canvas = self.canvas
        ix = cell.ix
        iy = cell.iy
        
        cx1,cy1,cx2,cy2 = self.get_win_ullr_at_ixy_canvas((ix,iy))
        SlTrace.lg(f"{ix},{iy}: {cell} :{cx1},{cy1}, {cx2},{cy2} ", "display_cell")
        if cell.mtype==BrailleCell.MARK_UNMARKED:
            canv_item = canvas.create_rectangle(cx1,cy1,cx2,cy2,
                                    outline="light gray")
        else:
            canv_item = canvas.create_rectangle(cx1,cy1,cx2,cy2,
                                    fill="dark gray",
                                    outline="dark gray")
        cell.canv_items.append(canv_item)
        self.update()
        color = self.color_str(cell._color)
        if len(color) < 1 or color[0] not in BrailleCell.color_for_character:
            color = "black"
        if show_points:
            dot_size = 1            # Display cell points
            dot_radius = dot_size//2
            if dot_radius < 1:
                dot_radius = 1
                dot_size = 2
            for pt in cell.points:
                dx,dy = self.get_point_win(pt)
                x0 = dx-dot_radius
                y0 = dy+dot_size 
                x1 = dx+dot_radius 
                y1 = dy
                canv_item = canvas.create_oval(x0,y0,x1,y1,
                                                fill=color)
                cell.canv_items.append(canv_item)
                SlTrace.lg(f"canvas.create_oval({x0},{y0},{x1},{y1}, fill={color})", "aud_create")
            self.update()    # So we can see it now 
            return
            
        dots = cell.dots
        cell_width = cx2-cx1
        cell_height = cy2-cy1       # y increases down
        # Fractional offsets from lower left corner
        # of cell rectangle
        ll_x = cx1      # Lower left corner
        ll_y = cy1
        ox1 = ox2 = ox3 = .3 
        ox4 = ox5 = ox6 = .7
        oy1 = oy4 = .15
        oy2 = oy5 = .45
        oy3 = oy6 = .73
        dot_size = .25*cell_width   # dot size fraction
        dot_radius = dot_size//2
        dot_offset = {1: (ox1,oy1), 4: (ox4,oy4),
                      2: (ox2,oy2), 5: (ox5,oy5),
                      3: (ox3,oy3), 6: (ox6,oy6),
                      }
        for dot in dots:
            offsets = dot_offset[dot]
            off_x_f, off_y_f = offsets
            dx = ll_x + off_x_f*cell_width
            dy = ll_y + off_y_f*cell_height
            x0 = dx-dot_radius
            y0 = dy+dot_size 
            x1 = dx+dot_radius 
            y1 = dy
            canv_item = canvas.create_oval(x0,y0,x1,y1,
                                            fill=color)
            cell.canv_items.append(canv_item) 
            SlTrace.lg(f"canvas.create_oval({x0},{y0},{x1},{y1}, fill={color})", "aud_create")
        self.update()
        pass
                
    def get_cell_center_win(self, ix, iy):
        """ Get cell's window rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: window(0-max): (xc,yc) where
            xc,yc are x,y coordinates of cell's center
        """
        x1,y1, x2,y2 = self.get_cell_rect_win(ix,iy)
        return (x1+x2)/2, (y1+y2)/2 
                
    def get_cell_rect_win(self, ix, iy):
        """ Get cell's window rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: window(0-max): (x1,y1,x2,y2) where
            x1,y1 are lower left coordinates
            x2,y2 are upper right coordinates
        """
        win_x1,win_y1,win_x2,win_y2 = self.get_win_ullr_at_ixy((ix,iy))
        return (win_x1,win_y1,win_x2,win_y2)
        
                    
        
    def get_cell_rect_tur(self, ix, iy):
        """ Get cell's turtle rectangle x, y  upper left, x,  y lower right
        :ix: cell x index
        :iy: cell's  y index
        :returns: (min_x,max_y, max_x,min_y)
        """
        if ix is None or ix < 0:
            SlTrace.lg(f"ix:{ix} < 0")
            ix = 0
        max_ix = len(self.cell_xs)-1
        if ix+1 > max_ix:
            SlTrace.lg(f"ix:{ix+1} >= {len(self.cell_xs)}")
            ix = max_ix-1
        if iy is None or iy < 0:
            SlTrace.lg(f"iy:{iy} < 0", "aud_move")
            iy = 0
        max_iy = len(self.cell_ys)-1
        if iy+1 > max_iy:
            SlTrace.lg(f"iy:{iy+1} >= {len(self.cell_ys)}")
            iy = max_iy-1
        x1 = self.cell_xs[ix]
        x2 = self.cell_xs[ix+1]
        y1 = self.cell_ys[iy+1]
        y2 = self.cell_ys[iy]
        return (x1,y1,x2,y2)
        
        
    def get_ixy_at(self, pt=None):
        """ Get cell(indexes) in which point resides
        If on an edge returns lower cell
        If on a corner returns lowest cell
        :pt: x,y pair location
                        in win coordinates
                default: current location
        :returns: ix,iy cell pair
                if out of bounds limit to min/max of ix,iy
        """
        if pt is None:
            pt = self.get_xy()
        x,y = pt
        x_min = self.get_x_min()
        y_min = self.get_y_min()
        ix = int((x-x_min)/self.win_width*self.grid_width)
        iy = int((y-y_min)/self.win_height*self.grid_height)
        #ix = int((x)/self.win_width*self.grid_width)        # TFD
        #iy = int((y)/self.win_height*self.grid_height)      # TFD
        return (ix,iy)

    def get_cells(self):
        """ Get cell dictionary (by (ix,iy)
        """
        return self.cells
    
    def get_cell_at(self, pt=None):
        """ Get cell at location, if one
        :pt: x,y pair location in turtle coordinates
                default: current location
        :returns: cell if at, else None
        """
        cell_ixy = self.get_ixy_at(pt)
        if cell_ixy is None:
            return None 
        
        if cell_ixy in self.cells:
            return self.cells[cell_ixy]
        
        return None

    def set_grid_path(self):
        self.grid_path = GridPath(self)
        
    def get_grid_path(self):
        return self.grid_path
    
    def get_win_ullr_at_ixy(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom)
        """
        ix,iy = ixy
        x_min = self.get_x_min()
        y_min = self.get_y_min()
        x_left = int(ix*(self.win_width/self.grid_width) + x_min)
        x_right = int((ix+1)*(self.win_width/self.grid_width) + x_min)
        y_top = int((iy*(self.win_height/self.grid_height) + y_min))
        y_bottom = int((iy+1)*(self.win_height/self.grid_height) + y_min)
        return (x_left,y_top, x_right,y_bottom)

    def get_win_ullr_at_ixy_canvas(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom) canvas coords
        """
        x_l,y_t, x_r,y_b = self.get_win_ullr_at_ixy(ixy)
        x_left,y_top = self.win_to_canvas((x_l,y_t))
        x_right,y_bottom = self.win_to_canvas((x_r,y_b))
        return (x_left,y_top, x_right,y_bottom)
        
    def win_to_canvas(self, xy):
        """ Convert window coordinates to canvas (min to max) to (0,max)
        :xy: x,y tuple
        :returns: x,y canvas tuple
        """
        x,y = xy
        canvas_x, canvas_y = (x-self.get_x_min(), y-self.get_y_min())
        return (canvas_x,canvas_y)
            
    def get_cell_at_ixy(self,cell_ixy, cells=None, ):
        """ Get cell at (ix,iy), if one
        :cell_ixy: (ix,iy)
        :cells: dictionary of cells by (ix,iy)
                default: self.get_cells()
        :returns: BrailleCell if one, else None
        """
        if cells is None:
            cells = self.get_cells()
        if cell_ixy in cells:
            return cells[cell_ixy]
        
        return None
        

    def is_at_cell(self, pt=None):
        """ Check if at cell
        :pt: x,y pair location in turtle coordinates
                default: current location
        :returns: True if at cell, else False
        """
        if self.get_cell_at(pt) is None:
            return False 
        
        return True

    def is_space(self, ixy, cells=None):
        """ Are we at a space
        :ixy: cell ix,iy indexes
        :cells: cells to check
                default: our cells - get_cells()
        :returns: True if a space (not a cell) 
        """
        if cells is None:
            cells = self.get_cells()
        cell = self.get_cell_at_ixy(cell_ixy=ixy, cells=cells)
        return True if cell is None else False
        
    def set_cell(self, pt=None, color=None):
        """ Set cell at pt, else current cell
        If no cell at current location create cell
        :pt: at point
            default: current location
        :color: color default: current color
        """
        if not self.is_at_cell(pt):
            self.create_cell(color=color)
        self.change_cell(pt=pt, color=color)

    def change_cell(self, pt=None, color=None):
        """ Change cell at pt if None current pt
        :pt: location default: current
        :color: color default: current color
        :returns: updated cell
        """
        if color is None:
            color = self.get_color()
        cell = self.get_cell_at(pt=pt)
        if cell is None:
            cell = self.create_cell(pt=pt, color=color)
        if cell is not None:
            cell.color_cell(color=color)
            self.display_cell(cell)
        self.update()
        return cell
                
    def color_str(self, color=None):
        """ Return color string
        :color: color specification str or tuple
        """
        color_str = color
        if (color_str is None
             or (isinstance(color_str, tuple)
                  and len(color_str) == 0)
             ):
            color_str = self.get_color()
        if isinstance(color_str,tuple):
            if len(color_str) == 1:
                color_str = color_str[0]
            else:
                color_str = "pink"  # TBD - color tuple work
        return color_str

    def is_visible(self):
        """ Check on visible mode
        """
        return self._visible
    
    def set_visible(self, val=True):
        """ Set cells visible/invisible
        Useful to give sighted a vision
        :val: set visible Default: True
        """
        SlTrace.lg(f"set_visible:{val}")
        self._visible = val
        if not val:
            self.set_show_marked(False)     # Make marked also invisible
        for cell in self.cells.values():
            self.set_visible_cell(cell, val)
                
    def set_visible_cell(self, cell, val=True):
        """ Set cells visible/invisible
        Useful to give sighted a vision
        :cell: figure cell
        :val: set visible Default: True
        """
        self.fte.set_visible_cell(cell, val=val)

    def announce_can_not_do(self, msg=None, val=None):
        """ Announce we can't do something
        """
        audio_beep = self.get_audio_beep()
        if audio_beep:
            audio_beep.announce_can_not_do(msg=msg, val=val)
        else:
            self.say_text(msg)
    
    def braille_for_color(self, color):
        """ Return dot list for color
        :color: color string or tuple
        :returns: list of dots 1,2,..6 for first
                letter of color
        """
        
        if color is None:
            color = self.get_color()
        if color is None:
            color = ("black")
        color = BrailleCell.color_str(color)
        c = color[0]
        dots = BrailleCell.braille_for_letter(c)
        return dots
    
    def braille_for_letter(self, c):
        """ convert letter to dot number seq
        :c: character
        :returns: dots tupple (1,2,3,4,5,6)
        """
        return BrailleCell.braille_for_letter(c)

    def clear_cells(self):
        """ remove all cells - clearing board
        Initialize empty board
        """
        cells = self.get_cells()
        for cell in list(cells.values()):
            self.erase_cell(cell)
        del self.cells
        self.cells = {}
        self.draw_cells(cells=self.cells)

        
        
        
    def create_cell(self, cell_ixy=None, color=None,
                    show=True):
        """ Create new cell ad cell_xy
        :cell_xy: ix,iy tuple default: current location
        :color: color default: curren color
        :show: show cell default:True
        :returns: cell
        """
        if cell_ixy is None:
            cell_ixy = self.get_ixy_at()
        if color is None:
            color = self.get_color()
        dots = self.braille_for_color(color)
        bc = BrailleCell(ix=cell_ixy[0], iy=cell_ixy[1],
                         dots=dots, color=color)
        if cell_ixy in self.cells:
            del self.cells[cell_ixy]
        self.cells[cell_ixy] = bc
        if show:
            self.update()
        return bc
            
    def complete_cell(self, cell, color=None):
        """ create/Fill braille cell
            Currently just fill with color letter (ROYGBIV)
        :cell: (ix,iy) cell index or BrailleCell
        :color: cell color default: current color
        :returns: created/modified cell
        """
        if color is None:
            color = self.get_color()
        dots = self.braille_for_color(color)
        bc = BrailleCell(ix=cell[0],iy=cell[1], dots=dots, color=color)
        self.cells[cell] = bc
        return bc

    def update(self):
        """ Update display
        """
        self.mw.update()

    def mainloop(self):
        self.mw.mainloop()
        
    def update_idle(self):
        """ Update pending
        """
        self.mw.update_idletasks()

    """
         Links to front end functions
    """

    def is_inbounds(self,ix=None, iy=None):
        """ Test if inbounds
        :ix: x index - default: don't test
        :iy: y index - default: don't test
        :returns: True iff in bounds
        """
        return self.fte.is_inbounds(ix=ix, iy=iy)

    def set_show_marked(self,val=True):
        """ Show marked "invisible" cells
        """
        self.fte.set_show_marked(val=val)

    def wait_on_output(self):
        """ Wait till queued output speech/tones completed
        """
        self.fte.wait_on_output()
        
    def win_print(self,*args, dup_stdout=False, **kwargs):
        """ print to listing area
        :*args: print-like args
        :**kwargs: print-flags
        :dup_stdout:  send duplicate to stdout
        """
        self.fte.win_print(args=args, dup_stdout=dup_stdout, kwargs=kwargs)

    def get_audio_beep(self):
        """ Get access to audio beep
        """
        return self.fte.get_audio_beep()

    def cursor_update(self):
        """ Update cursor (current position) display
        """
        self.fte.cursor_update()
        
    def mark_cell(self, cell,
                  mtype=BrailleCell.MARK_SELECTED):
        """ Mark cell for viewing of history
        :cell: Cell(BrailleCell) or (ix,iy) of cell
        :mtype: Mark type default: MARK_SELECTED
        """
        self.fte.mark_cell(cell=cell, mtype=mtype)

    def set_color(self, color):
        """ Set current color
        """
        self.fte.set_color(color=color)
        
    def get_color(self):
        """ Get current color
        """
        return self.fte.get_color()
        
    def set_drawing(self, val=True):        
        self.fte.set_drawing(val=val)
        return val
    
    def is_drawing(self):
        return self.fte.is_drawing()

        
    def key_goto(self):
        """ Go to closest figure
            go to figure not in one already
            else go one step within current figure, toward
            longest inside path
        """
        self.fte.key_goto()

    def set_scanning(self, cells=None):
        self.fte.set_scanning(cells=cells)
        
    def get_vol(self, ix, iy, eye_ixy_l=None, eye_ixy_r=None):
        """ Get tone volume for cell at ix,iy
        volume(left,right)
        Volume(in decibel): k1*(k2-distance from eye), k1=1, k2=0
        :ix: cell ix
        :iy: cell iy
        :eye_xy_l: left eye/ear at x,y default: self.eye_xy_l
        :eye_xy_r: right eye/ear at x,y  default: self.eye_xy_r
        return: volume(left,right) in decibel
        """
        return self.fte.get_vol(ix=ix, iy=iy, eye_ixy_l=eye_ixy_l, eye_ixy_r=eye_ixy_r)

    def set_xy(self, xy):
        """ Set our internal win x,y
        :xy: x,y tuple as new win coordinates
        :returs: (x,y) tuple
        """
        return self.fte.set_xy(xy)

    def get_xy(self):
        """ Get current xy pair
        :returns: (x,y) tuple
        """
        return self.fte.get_xy() 

    def get_xy_canvas(self, xy=None):
        """ Get current xy pair on canvas (0-max)
        :xy: xy tuple default: current xy
        :returns: (x,y) tuple
        """
        return self.fte.get_xy_canvas(xy=xy)
    
    def set_x_min(self, val):
        """ Set min
        :val: new x_min
        :returns: new x_min
        """
        return self.fte.set_x_min(val)

    def get_x_min(self):
        return self.fte.get_x_min()

    def set_x_max(self, val):
        """ Set max
        :val: new x_max
        :returns: new x_min
        """
        return self.fte.set_x_max(val)

    def get_x_max(self):
        return self.fte.get_x_max()
        
 
    def set_y_min(self, val):
        """ Set min
        :val: new x_min
        :returns: new x_min
        """
        return self.fte.set_y_min(val)

    def get_y_min(self):
        return self.fte.get_y_min()

    def set_y_max(self, val):
        """ Set max
        :val: new y_max
        :returns: new x_min
        """
        return self.fte.set_y_max(val)

    def get_y_max(self):
        return self.fte.get_x_max()

                
    def pos_check(self, x=None, y=None, force_output=False, with_voice=False):
        """ Do position checking followed by report queue processing
        :x: x postion default: current
        :y: y position default: current
        :force_output: force output, flushing current queue
        :with_voice: say if, even if beep enable
        """
        self.fte.pos_check(x=x, y=y, force_output=force_output,
                             with_voice=with_voice)

    def is_silent(self):
        return self.fte.is_silent()
if __name__ == "__main__":
    SlTrace.clearFlags()
    menu_str = "n:sh;d:d"
    key_str = (
                "d"
                ";c;g;9;9;9;9"
                ";c;r;7;7;7;7"
                ";c;v;2;2;2;c;r;2;2;c;o;2;2;2"
                  ";u;8;8;8;8;8;8;8;8;d"
                ";c;o;1;1;1;1"
                ";c;b;3;3;3;3"
                ";c;g;6;6;c;i;6;6;6;c;v;6;6;6"
                ";w"
            )
    aw = AudioDrawWindow(title="AudioDrawWindow Self-Test",
                         menu_str=menu_str,
                         key_str=key_str)

    aw.fte.do_menu_str("n:nu;d:s")
    aw.fte.do_key_str("u")
    
    
    
    
    aw.mw.mainloop()
    



# show_square_loop_colors_w80.py  # standard
# Display a square with colored sides

from turtle import *		     # Get standard stuff

colors = ["red","orange","yellow","green"]

for colr in colors:
    width(80)
    color(colr)
    forward(200)
    right(90)
done()		    # Complete drawings



#wx_adw_display_pending.py  10Mar2024  crs  Author
import wx

from select_trace import SlTrace

"""
AudioDrawWindow display pending
Facilitate the construction of items to display and their display
Items of interest
    1. BrailleCells, initial and traversed
    2. Large lists of BrailleCells
    3. Cursor, indicator of current traversal position inside cell, outside cell
    4. Magnification Selection - rectangle surrounding region to magnify

User sets function to do actual item display
by calling set_display_fun(<user's item display function>)    
"""
            
class AdwDisplayPending:
    def __init__(self, canv_pan):
        """ pending display items
        :canv_pan: canvas panel (CanvPanel)
        """
        self.canv_pan = canv_pan
        self.disp_fun = None    # Set as user's item displaying function
        self.clear()

    def clear(self):
        """ Clear/initialize pending
        """
        self.items = []         # one-time list for update
        self.perm_items = {}    # permanent dictionary, by id, for redrawing
        self.prev_npending = None   # Track changes
                
    
    def add_item(self, item):
        """ Add display item
        :item: item/id (CanvasPanelItem)
        """
        if type(item) == int:
            itm = self.canv_pan.id_to_item(item)
        else:
            itm = item
        if type(itm) == int:
            SlTrace.lg("item: {item}")
        self.items.append(itm)
        self.add_perm_item(itm)
        ##self.canv_pan.refresh_item(itm)
        
    def add_perm_item(self, item):
        """ Add permanent item to support redrawing
        TBD handling item changes, overlay etc.
        
        :item: item to be added for redrawing
        """
        self.perm_items[item.canv_id] = item
        
    def add_cell(self, di_item):
        """ Add cell to be displayed
        :cell: display item (AdwDisplayPendingItem)
        """            
        self.add_item(di_item)
        
    def add_cursor(self, cursor):
        """ Add cursor to be displayed
        :cursor: item
        """            
        self.add_item(cursor)
        
    def add_mag_sel(self, mag_sel):
        """ Add magnification selection to be displayed
        :mag_sel: magnification selection
        """            
        self.add_item(mag_sel)
        
    def add_cpan_item(self, cpan_item):
        """ Add cell to be displayed
        :cpan_item: CanvasPanelItem item to display
        """            
        self.add_item(cpan_item)

    def get_displayed_items(self):
        """ Get list of displayed items
        :returns: list of permanently displayed values (AdwDisplayPendingItem)
        """
        return self.perm_items.values()

    def is_overlapping(self, item1, item2):
        """ Check if two display items are overlapping
        :returns: True iff overlapping
        """
    def clear_pending(self):
        """ Clear pending
        """
        self.clear()    
        
    def display_pending(self, dc):
        """ Display list and clear it
        :dc: wx.PaintDC(self)
        """
        if len(self.items) > 0:
            self.npending = len(self.items)
            
            color = self.canv_pan.color
            #dc = wx.PaintDC(self.canv_pan.grid_panel)
            dc.SetPen(wx.Pen(color))
            style = wx.SOLID
            dc.SetBrush(wx.Brush(color, style))
            if self.prev_npending is None or self.npending != self.prev_npending:
                SlTrace.lg(f"{self.npending} display_pending prev = {self.prev_npending}",
                           "display_pending")
                self.prev_npending = self.npending
            create_cursor = None
            create_mag_select = None
            for diitem in self.items:
                if diitem.canv_type == "create_cursor":
                    SlTrace.lg(f"diitem:{diitem}", "display_pending")
                    create_cursor = diitem
                elif diitem.canv_type == "create_mag_select":
                    create_mag_select = diitem
                else:    
                    self.display_item(diitem)
            if create_cursor is not None:
                self.display_item(create_cursor)
            #self.items = []     # Clear list
    
    def display_item(self, item):
        """ Display item
        :diitem: DisplayListItem item to display
        """
        if self.disp_fun:
            self.disp_fun(item)
            return
        self.draw_item(item)

    def draw_item(self, item):
        """ Draw item
        :item: CanvasPanalItem/canv_id item to draw
        """
        self.canv_pan.draw_item(item)
            
    def set_display_item_fun(self, disp_fun):
        """ Set display item function
        :disp_fun: display function for dispaly item
        """
        self.disp_fun = disp_fun
        


# wx_rpc.py 30Aug2024  crs  from examples/rpc.py 
# rpc.py    27Aug2024  from https://github.com/TarasZhere/RPC

import json
import socket
import inspect
import traceback
from threading import Thread
from select_trace import SlTrace
SIZE = 2**16


class RPCServer:

    def __init__(self, host:str='0.0.0.0', port:int=8080) -> None:
        self.host = host
        self.port = port
        self.address = (host, port)
        self._methods = {}

    def help(self) -> None:
        SlTrace.lg('REGISTERED METHODS:')
        for method in self._methods.items():
            SlTrace.lg('\t',method)

    '''

        registerFunction: pass a method to register all its methods and attributes so they can be used by the client via rpcs
            Arguments:
            instance -> a class object
    '''
    def registerMethod(self, function) -> None:
        try:
            self._methods.update({function.__name__ : function})
        except:
            raise Exception('A non method object has been passed into RPCServer.registerMethod(self, function)')

    '''
        registerInstance: pass a instance of a class to register all its methods and attributes so they can be used by the client via rpcs
            Arguments:
            instance -> a class object
    '''
    def registerInstance(self, instance=None) -> None:
        try:
            # Regestring the instance's methods
            for functionName, function in inspect.getmembers(instance, predicate=inspect.ismethod):
                if not functionName.startswith('__'):
                    self._methods.update({functionName: function})
        except:
            raise Exception('A non class object has been passed into RPCServer.registerInstance(self, instance)')

    '''
        handle: pass client connection and it's address to perform requests between client and server (recorded fucntions or) 
        Arguments:
        client -> 
    '''
    def __handle__(self, client:socket.socket, address:tuple):
        SlTrace.lg(f'Managing requests from {address}.', "rpc")
        while True:
            try:
                functionName, args, kwargs = json.loads(client.recv(SIZE).decode())
            except: 
                SlTrace.lg(f'! Client {address} disconnected.')
                break
            # Showing request Type
            SlTrace.lg(f'> {address} : {functionName}({args})', "rpc")
            
            try:
                response = self._methods[functionName](*args, **kwargs)
            except Exception as e:
                estr = "Exception:" + str(e)
                tbstk = traceback.extract_stack()
                tbstk_lst = traceback.format_list(tbstk)
                tbstr = "\n".join(tbstk_lst)
                SlTrace.lg(f"RPCServer Exception: {estr}")
                SlTrace.lg(f"__handle__({functionName}, {args}, {kwargs})")
                SlTrace.lg(f"{tbstr}\n")
                # Send back exeption if function called by client is not registred 
                client.sendall(json.dumps(estr).encode())
            else:
                client.sendall(json.dumps(response).encode())


        SlTrace.lg(f'Completed request from {address}.')
        client.close()
    
    def run(self) -> None:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.bind(self.address)
            sock.listen()

            SlTrace.lg(f'+ Server {self.address} running', "rpc")
            while True:
                try:
                    client, address = sock.accept()

                    Thread(target=self.__handle__, args=[client, address]).start()

                except KeyboardInterrupt:
                    SlTrace.lg(f'- Server {self.address} interrupted')
                    break



class RPCClient:
    def __init__(self, host:str='localhost', port:int=8080) -> None:
        self.__sock = None
        self.__address = (host, port)


    def isConnected(self):
        try:
            self.__sock.sendall(b'test')
            self.__sock.recv(SIZE)
            return True

        except:
            return False


    def connect(self):
        try:
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            SlTrace.lg(f"connect({self.__address})", "rpc")
            self.__sock.connect(self.__address)
        except EOFError as e:
            SlTrace.lg(e)
            SlTrace.lg(f"connect({self.__address})")
            raise Exception('Client was not able to connect.')
    
    def disconnect(self):
        try:
            self.__sock.close()
        except:
            pass


    def __getattr__(self, __name: str):
        def excecute(*args, **kwargs):
            self.__sock.sendall(json.dumps((__name, args, kwargs)).encode())

            response = json.loads(self.__sock.recv(SIZE).decode())
   
            return response
        
        return excecute

    def __del__(self):
        try:
            self.__sock.close()
        except:
            pass


# adw_menus    09Mar2023  crs, Split off from adw_front_end

import wx

from select_trace import SlTrace
from wx_trace_control_window import TraceControlWindow


class MenuDisp:
    """ Menu dispatch table entry
    Supporting multiple mode dispatch (e.g, Dropdown item plus command mode)
    """

    def __init__(self, label, command, underline):
        self.label = label
        self.command = command
        self.underline = underline
        self.shortcut = label[underline].lower()
        self.Properties = None


class AdwMenus:
    def __init__(self, adw_front_end, frame=None):
        """ Setup menus for AudioDrawWindow
        :adw_front_end: (AdwFrontEnd)
        :frame: frame containing menus
        """
        self.fte = adw_front_end
        if frame is None:
            frame = wx.Frame(None)
        self.frame = frame
        self.frame.Show()
        
        self.menu_setup()

    def on_alt_a(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_a")

    def on_alt_m(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_m")

    def on_alt_n(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_n")

    def on_alt_f(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_f")

    def on_alt_s(self, _=None):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_s")

    def menu_setup(self):
        """ Setup menu system
        """
        # creating a menu instance
        menubar = wx.MenuBar()
        menu_name_list = [
            "file", "magnify","navigate",
            "draw", "scanning", "auxiliary"
            ]
            
        # Settings for each menu    
        menus_settings = {
            "file" : {"heading" : "File"},
            "magnify" : {"heading" : "Magnify"},
            "navigate" : {"heading" : "Navigate"},
            "draw" : {"heading" : "Draw"},
            "scanning" : {"heading" : "Scanning"},
            "auxiliary" : {"heading" : "Auxiliary"},
        }
        
        """ menus_items defines all the menu drop downs
        SEP - specifies a separator line
        
        "name" specifies name and heading
        if value is a str the heading == name
        else name is a tuple with
            name = tuple[0]
            heading = tuple[1]
        """
        
        SEP = {"sep": "sep"}    # Menu item separator
          
        menus_items = {
            "file":
                [
                {"name" : "Open",  "cmd" : self.TBD},
                {"name" : "Save",  "cmd" : self.TBD},
                SEP,
                {"name" : "Log",   "cmd" : self.LogFile},
                {"name" : "Properties",   "cmd" : self.PropertiesFile},
                SEP,
                {"name" : "E&xit",  "cmd" : self.pgm_exit},
                ],
            "magnify" :
                [
                {"name" : "Help", "cmd" : self.mag_help},
                {"name" : ("rem_pos","Remove Pos History"), "cmd" : self.erase_pos_history},
                {"name" : "Select", "cmd" : self.mag_select},
                {"name" : "Expand To &Fill", "cmd" : self.mag_expand_to_fill},
                {"name" : "Expand &Top", "cmd" : self.mag_expand_top},
                {"name" : "View", "cmd" : self.mag_view_cmd},              
                ],
            "navigate" :
                [
                {"name" : ("help","Help"), "cmd" : self.nav_help},
                {"name" : ("add_loc","add At loc"), "cmd" : self.nav_add_loc},
                {"name" : ("no_add_loc","b-remove At loc"), "cmd" : self.nav_no_add_loc},
                {"name" : ("echo on","echo input on"), "cmd" : self.nav_echo_on},
                {"name" : ("hd","echo &off"), "cmd" : self.nav_echo_off},
                {"name" : ("hd","visible cells"), "cmd" : self.nav_make_visible},
                {"name" : ("hd","invisible cells"), "cmd" : self.nav_make_invisible},
                {"name" : ("hd","marked"), "cmd" : self.nav_show_marked},
                {"name" : ("hd","noisy"), "cmd" : self.make_noisy},
                {"name" : ("hd","silent"), "cmd" : self.make_silent},
                {"name" : ("hd","talking"), "cmd" : self.nav_make_talk},
                {"name" : ("hd","log talk"), "cmd" : self.nav_logt},
                {"name" : ("hd","no log tal&k"), "cmd" : self.nav_no_logt},
                {"name" : ("hd","position"), "cmd" : self.nav_say_position},
                {"name" : ("hd","redraw figure"), "cmd" : self.nav_redraw},
                {"name" : ("hd","a&udio beep"), "cmd" : self.nav_audio_beep},
                {"name" : ("hd","q no audio beep"), "cmd" : self.nav_no_audio_beep},
                {"name" : ("hd","x enable mouse navigation"), "cmd" : self.nav_enable_mouse},
                {"name" : ("hd","y disable mouse navigation"), "cmd" : self.nav_disable_mouse},
                ],
            "draw" :
                [
                {"name" : "Help", "cmd" : self.draw_help},
                {"name" : "drawing", "cmd" : self.start_drawing},
                {"name" : "stop_drawing", "cmd" : self.stop_drawing},
                {"name" : "count_cells", "cmd" : self.count_cells},
                ],
            "scanning" :
                [
                {"name" : "Help", "cmd" : self.scan_help},
                {"name" : "c -  combine wave - faster scan", "cmd" : self.scan_combine_wave},
                {"name" : "d - disable combine wave", "cmd" : self.scan_disable_combine_wave},
                {"name" : "f - flip skip space", "cmd" : self.flip_skip_space},
                {"name" : "r - flip skip run", "cmd" : self.flip_skip_run},
                {"name" : "Start scanning", "cmd" : self.start_scanning},
                {"name" : "S&top scanning", "cmd" : self.stop_scanning},
                {"name" : "No item wait", "cmd" : self.scan_no_item_wait},
                {"name" : "Wait for item", "cmd" : self.scan_item_wait},
                ],
            "auxiliary" :
                [
                {"name" : "Trace", "cmd" : self.trace_menu},
                ],
            
        }
        
        self.frame.SetMenuBar(menubar)
        self.menus_cmd_menu_item = {}   # by menu short cut by cmd short cut
                                        # m[menu_sc][menu_item_sc] = cmd
        
        for menu_name in menu_name_list:
            menu_settings = menus_settings[menu_name]
            menu_heading = menu_settings["heading"]
            if "&" not in menu_heading:
                menu_heading = "&" + menu_heading
            menu_sci = menu_heading.find("&")
            menu_sc = menu_heading[menu_sci+1].lower()
            if menu_sc in self.menus_cmd_menu_item:     # Is shortcut in use ?
                raise Exception(f"Duplicate menu shortcut {menu_sc} for"
                        f" menu {menu_heading}")
            else:
                menu_items_scs = self.menus_cmd_menu_item[menu_sc] = {}
            menu = wx.Menu()
            menubar.Append(menu,  menu_heading)
            menu_items = menus_items[menu_name]
            for menu_item_specs in menu_items:
                if "sep" in menu_item_specs:
                    menu_item = menu.Append(wx.ID_SEPARATOR)
                else:
                    menu_cmd = menu_item_specs["cmd"]
                    name_heading = menu_item_specs["name"]
                    if isinstance(name_heading,str):
                        menu_item_heading = menu_item_name = name_heading
                    elif len(name_heading)==1:
                        menu_item_name = menu_item_heading = name_heading[0]
                    elif len(name_heading) == 2:
                        menu_item_name = name_heading[0]
                        menu_item_heading = name_heading[1]
                    if "&" not in menu_item_heading:
                        menu_item_heading = "&" + menu_item_heading
                    menui_sci = menu_item_heading.find("&")
                    menui_sc = menu_item_heading[menui_sci+1].lower()
                    if menui_sc in menu_items_scs:
                        raise Exception(f"Duplicate menu {menu_heading} item {menu_item_heading}"
                                        f" \nshortcut {menui_sc}"
                                f" \nshortcuts in use {list(menu_items_scs)}")
                    else:
                        menu_items_scs[menui_sc] = menu_cmd
   
                    menu_item = menu.Append(wx.ID_ANY, menu_item_heading)
                self.frame.Bind(wx.EVT_MENU, menu_cmd, menu_item)

    def get_menu_cmd(self, menu_sc, mi_sc):
        """ get menu cmd
        :menu_cs: menu shortcut case insensitive
        :mi_cs: menu item shortcut case insensitive
        :returns: menu cmd, if none - None
        """
        menu_sc = menu_sc.lower()
        if menu_sc not in self.menus_cmd_menu_item:
            return None # no menu shortcut
        mcmis = self.menus_cmd_menu_item[menu_sc]
        if mi_sc not in mcmis:
            return None
        
        return mcmis[mi_sc]

    def get_menu_scs(self):
        """ Get list of menu short cuts
        :returns: list of menu shortcuts
        """
        menu_scs = list(self.menus_cmd_menu_item)
        return menu_scs

    def get_menu_item_scs(self, menu_sc):
        """ Get list of menu item short cuts
        :menu_sc: menu shortcut
        :returns: list of menu itme shortcuts
        """
        if menu_sc not in self.menus_cmd_menu_item:
            return None
        
        menu_items = self.menus_cmd_menu_item[menu_sc]
        return list(menu_items) 
              
        
    def file_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.file_menu.add_command(label=label, command=command,
                                   underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.file_dispatch[menu_de.shortcut] = menu_de

    def TBD(self, _=None):
        SlTrace.lg("To be developed")

    def LogFile(self, _=None):
        SlTrace.lg("Show LogFile")

    def PropertiesFile(self, _=None):
        SlTrace.lg("Show PropertiesFile")
                
    def file_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.file_dispatch:
            raise Exception(f"file option:{short_cut} not recognized")
        menu_de = self.file_dispatch[short_cut]
        menu_de.command()

    def File_Open_tbd(self):
        print("File_Open_menu to be determined")

    def File_Save_tbd(self):
        print("File_Save_menu to be determined")


    def draw_help(self, _=None):
        """ Help for drawing
        """
        """ Help - list command (Alt-d) commands
        """
        help_str = """
        Help - list drawing setup commands (Alt-d) commands
        h - say this help message
        d - Start/enable drawing
        s - stop/disable drawing
        c - count cells (squares)
        Escape - flush pending report output
        """
        self.speak_text(help_str)

    """ Magnify support package
    """

    """ Magnify menu commands  """

    def mag_help(self, _=None):
        """ Help for Alt-m commands
        """
        """ Help - list command (Alt-m) commands
        """
        help_str = """
        Help - list magnify commands (Alt-m) commands
        h - say this help message
        l - expand magnify selected region left/right
        s - select/mark magnify region
        u - expand magnify selected region up/down top/bottom
        v - view region (make new window)
        """
        self.speak_text(help_str)

    """ End of Magnify support
    """

    """ Scanning support package
    """

    """ Scanning menu commands  """

    def scan_help(self, _=None):
        """ Help for Alt-s commands
        """
        """ Help - list command (Alt-s) commands
        """
        help_str = """
        Help - list scanning commands (Alt-s) commands
        h - say this help message
        c - combine wave - faster scan
        d - disable combine wave - viewing in window
        k - toggle skip space
        s - Start scanning mode
        t - Stop scanning mode
        n - no_item_wait
        w - wait for items
        """
        self.speak_text(help_str)

    def scan_combine_wave(self, _=None):
        self.set_combine_wave()

    def scan_disable_combine_wave(self, _=None):
        self.set_combine_wave(val=False)

    def scan_no_item_wait(self, _=None):
        self.set_no_item_wait()

    def scan_item_wait(self, _=None):
        """clear no item wait
        """
        self.set_no_item_wait(val=False)

    """ End of Scanning support
    """

    """ Navigate support package
    """

    def nav_help(self, _=None):
        """ Help for Alt-n commands
        """
        """ Help - list command (Alt-n) commands
        """
        help_str = """
        Help - list navigate commands (Alt-n) commands
        h - say this help message
        a - Start reporting position
        b - remove 
        z - stop reporting position
        e - echo input on
        o - echo off
        v - visible cells
        i - invisible cells
        r - redraw figure
        s - silent speech
        t - talking speech
        l - log speech
        m - show marked(even if invisible)
        n - no log speech
        p - report position
        u - audio beep
        d - no audio beep
         Escape - flush pending report output
        """
        self.speak_text(help_str)

    """
    Trace support
    """

    def trace_menu(self, _=None):
        TraceControlWindow()

    """
    Links to front end 
    """
    
    """
    General
    """

    def erase_pos_history(self, _=None):
        self.fte.erase_pos_history()

    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9
            
        """
        self.fte.speak_text(msg=msg, msg_type=msg_type,
                            dup_stdout=dup_stdout,
                            rate=rate, volume=volume)

    """
     File links
      most local
    """

    def pgm_exit(self, _=None):
        self.fte.pgm_exit()

    """
    Magnification links
    """

    def mag_select(self, _=None):
        """ Select magnification region
            -rectangle including all figure cells traveled so far
            :returns: True if some selected else False
        """
        return self.fte.mag_select()

    def mag_expand_to_fill(self, _=None):
        """ Expand selection region to enclose figure
        """
        self.fte.mag_expand_to_fill()

    def mag_expand_top(self, _=None):
        """ Expand selection region top/bottom by 20%
        """
        self.fte.mag_expand_top()

    def mag_view(self, cells=None, _=None):
        """ View selected region, creating a new AudioDrawWindow
        """
        self.fte.mag_view(cells=cells)

    def mag_view_cmd(self, _=None):
        """ View selected region (menu dispatch cmd)
        creating a new AudioDrawWindow menu cmd
        _ : event - not used
        """
        self.fte.mag_view()

    """ drawing
    """

    def start_drawing(self, _=None):
        """ Start/enable drawing
        """
        self.fte.start_drawing()

    def stop_drawing(self, _=None):
        """ Stop/disable drawing
        """
        self.fte.stop_drawing()

    def count_cells(self, _=None):
        """ report number of cells in figure
        """
        ###???self.fte.count_cells()
    
    """ End of Magnification links """

    """
     Navigation / Front End links
     """

    def set_enable_mouse(self, val=True):
        """ Enable/Disable mouse dragging operation
        :val: value to set default: True - enable
        """
        self.fte.set_enable_mouse(val=val)

    def is_enable_mouse(self):
        """ Check if mouse dragging enabled
        """
        return self.fte.is_enable_mouse()

    def nav_add_loc(self, _=None):
        self.fte.nav_add_loc()

    def nav_no_add_loc(self, _=None):
        self.fte.nav_no_add_loc()

    def nav_echo_on(self, _=None):
        self.fte.nav_echo_on()

    def nav_echo_off(self, _=None):
        self.fte.nav_echo_off()

    def nav_make_visible(self, _=None):
        self.fte.nav_make_visible()

    def nav_make_invisible(self, _=None):
        self.fte.nav_make_invisible()

    def nav_show_marked(self, _=None):
        self.fte.nav_show_marked()

    def make_noisy(self, _=None):
        self.fte.make_noisy()

    def make_silent(self, _=None):
        self.fte.make_silent()

    def nav_make_talk(self, _=None):
        self.fte.nav_make_talk()

    def nav_logt(self, _=None):
        self.fte.nav_logt()

    def nav_no_logt(self, _=None):
        self.fte.nav_no_logt()

    def nav_say_position(self, _=None):
        self.fte.nav_say_position()

    def nav_redraw(self, _=None):
        self.fte.nav_redraw()

    def nav_audio_beep(self, _=None):
        self.fte.nav_audio_beep()

    def nav_no_audio_beep(self, _=None):
        self.fte.nav_no_audio_beep()

    def nav_enable_mouse(self, _=None):
        self.set_enable_mouse()

    def nav_disable_mouse(self, _=None):
        self.set_enable_mouse(False)

    """ End of Navigation links """

    """
    Scanning support
        through fte
    """

    def set_combine_wave(self, val=True):
        """ Enable/disable combine wave scanning mode
        :val: value for mode
        """
        self.fte.set_combine_wave(val=val)

    def set_no_item_wait(self, val=True):
        """ Set/clear scanning no_wait option
        :val: True - no waiting
        """
        self.fte.set_no_item_wait(val=val)

    def flip_skip_space(self, _=None):
        """ Flip skipping spaces
        """
        self.fte.flip_skip_space()

    def flip_skip_run(self, _=None):
        """ Flip skipping run of equals
        """
        self.fte.flip_skip_run()

    def start_scanning(self, _=None):
        self.fte.start_scanning()

    def stop_scanning(self, _=None):
        self.fte.stop_scanning()

"""
Used to stub out FTE
"""
class FteFake:
    def pgm_exit(self):
        SlTrace.lg("fte.pgm_exit()")
        import sys
        sys.exit()
        
    def speak_text(self, msg, msg_type=None,
                    dup_stdout=True, rate=None,
                    volume=None):
        SlTrace.lg(f"fte.speak_text({msg},"
                    f"msg_type={msg_type}msg)"
                    f")")

    """ mag support"""
    def erase_pos_history(self):
        SlTrace.lg("fte.erase_pos_history")
    def mag_select(self):
        SlTrace.lg("mag_select")    
    def mag_expand_to_fill(self):
        SlTrace.lg("mag_expand_to_fill")    
    def mag_exapnd_top(self):
        SlTrace.lg("mag_exapnd_top")    
    def mag_view(self):
        SlTrace.lg("mag_view")    

    """ nav support"""

    def nav_audio_beep(self, _=None):
        SlTrace.lg("fte.nav_audio_beep")

    def nav_no_audio_beep(self, _=None):
        SlTrace.lg("fte.nav_no_audio_beep")

    def nav_make_visible(self, _=None):
        SlTrace.lg("fte.nav_make_visible")

    def nav_make_invisible(self, _=None):
        SlTrace.lg("fte.nav_make_invisible")

    def set_enable_mouse(self, _=None,val=None):
        SlTrace.lg("fte.set_enable_mouse")

    def nav_show_marked(self, _=None):
        SlTrace.lg("fte.nav_show_marked")

    def nav_redraw(self, _=None):
        SlTrace.lg("fte.nav_redraw")

    def nav_say_position(self, _=None):
        SlTrace.lg("fte.nav_say_position")

    def nav_make_talk(self, _=None):
        SlTrace.lg("fte.nav_make_talk")

    def nav_logt(self, _=None):
        SlTrace.lg("fte.nav_logt")

    def nav_no_logt(self, _=None):
        SlTrace.lg("fte.nav_no_logt")

    def make_noisy(self, _=None):
        SlTrace.lg("fte.make_noisy")

    def make_silent(self, _=None):
        SlTrace.lg("fte.make_silent")
    
    def nav_add_loc(self):
        SlTrace.lg("fte.nav_add_loc()")
    
    def nav_no_add_loc(self):
        SlTrace.lg("fte.nav_no_add_loc()")
    
    def nav_echo_on(self):
        SlTrace.lg("fte.nav_echo_on()")
    
    def nav_echo_off(self):
        SlTrace.lg("fte.nav_echo_off()")
        
    def start_drawing(self):
        SlTrace.lg(f"fte.start_drawing()")

    def stop_drawing(self):
        SlTrace.lg(f"fte.stop_drawing() is this in fte?")
    
    def set_combine_wave(self, val=True):
        SlTrace.lg("fte.set_combine_wave")
    
    def  flip_skip_space(self):
        SlTrace.lg("fte.flip_skip_space")
    
    def  flip_skip_run(self):
        SlTrace.lg("fte.flip_skip_run")
    
    def start_scanning(self):
        SlTrace.lg("fte.start_scanning")
                                                                                                                                        
    def stop_scanning(self):
        SlTrace.lg("fte.stop_scanning") 

    def set_no_item_wait(self, val=True):
        SlTrace.lg(f"fte.set_no_item_wait(val={val})") 


if __name__ == "__main__":
    from wx_adw_menus import AdwMenus, FteFake
                  
    app = wx.App()
    frame = wx.Frame(None)
    fte = FteFake()
    menus = AdwMenus(fte, frame=frame)

    men_scs = menus.get_menu_scs()
    SlTrace.lg(f"Menus: {men_scs}")
    for men_sc in men_scs:
        SlTrace.lg(f"menu {men_sc}")    
        menu_item_scs = menus.get_menu_item_scs(men_sc)    
        SlTrace.lg(f"menu {men_sc} items: {menu_item_scs}")

    SlTrace.lg("Calling each function")
    for men_sc in men_scs:
        SlTrace.lg(f"menu {men_sc}")    
        men_item_scs = menus.get_menu_item_scs(men_sc)
        for menui_sc in men_item_scs:    
            cmd_command = menus.get_menu_cmd(men_sc, menui_sc)
            SlTrace.lg(f"Menu:{men_sc} Item:{menui_sc}", "menu")
            if men_sc == "f" and menui_sc.lower() == "x":
                SlTrace.lg(f"Skipping Menu:{men_sc} Item:{menui_sc}")
                continue
            cmd_command()
            
        
        
    app.MainLoop()


#wx_win.py    24May2023  crs, Author
"""
Reference to wx window support
"""
import time
import wx

from select_trace import SlTrace

class WxWin:
    def __init__(self, adw, title=None):
        """ Setup wxPython window access
        :adw: AudioDrawWindow base
        :title: optional window title
        """
        self.adw = adw
        self.app = wx.App()
        frame = wx.Frame(None,  title=title)

        adw_panel = wx.Panel(frame)
        adw_vbox = wx.BoxSizer(wx.VERTICAL)
        tout_hbox = wx.BoxSizer(wx.HORIZONTAL)
        adw_vbox.Add(tout_hbox, flag=wx.CENTER)
        tout_ctrl = wx.TextCtrl(adw_panel)
        tout_hbox.Add(tout_ctrl)
        self.tout_ctrl = tout_ctrl

        
        adw_panel.SetSizerAndFit(adw_vbox)
        frame.Show()

    def MainLoop(self):
        self.app.MainLoop()

    def update(self):
        time.sleep(.1)
        wx.Yield()
                        
if __name__ == "__main__":
    from wx_audio_draw_window import AudioDrawWindow
    
    #adw = AudioDrawWindow(setup_wx_win=False)
    adw = None
    wx_win = WxWin(adw, "WxWin Self Test")
    wx_win.MainLoop()



# wx_audio_beep.py  26Oct2023  crs, from audio_beep
#                   25Nov2022  crs,
""" 
Audio Support for audio_window using wxPython
"""
import winsound
from select_trace import SlTrace
from wx_sinewave_beep import SineWaveBeep

class AudioBeep:
    beep_dur = 200      # Default beep duration
    color_fm = 1000     # Color frequency multiplier
    other_freq = color_fm//4
    other_dur = beep_dur*2
    blank_freq = color_fm//3
    blank_dur = beep_dur*2
    
    color_freqs = {
        "OTHER" : other_freq,
        "BLANK" : blank_freq,
        "red" : 1*color_fm,
        "orange" : 2*color_fm,
        "yellow" : 3*color_fm,
        "green" : 4*color_fm,
        "blue" :  5*color_fm,
        "indigo" : 6*color_fm,
        "violet" : 7*color_fm,
        "black" : 8*color_fm,
        "white" : 9*color_fm,
        "gray" : 10*color_fm,
        }
    
    def __init__(self, awin, silence_checker=None):
        """ Setup audio sound support
        :awin: instance of AudioWindow
        :silence_checker: function to check if we should
                        go silent
        """
        self.awin = awin
        self._silence_checker = silence_checker
        self.sinewave_beep = SineWaveBeep(awin, silence_checker=silence_checker)
        self.has_sinewave = True        # TEMP - push checking deeper

    def announce_can_not_do(self, msg=None, val=None):
        """ Announce we can't do something
        """
        SlTrace.lg("announce_can_not_do()", "bounds")
        if self.has_sinewave:
            self.sinewave_beep.announce("can_not_do")
        else:    
            winsound.Beep(frequency=self.color_fm//5, duration=self.beep_dur//3)


    def silence(self):
        """ Check if we are silenced
        """
        if self._silence_checker is None:
            return False 
        
        if self._silence_checker():
            return True    # We're silent
       
        return False    # We make noise 
            
    def on_edge(self):
        if self.silence():
            return
        if self.has_sinewave:
            self.sinewave_beep.announce("on_edge")
        else:
            winsound.Beep(200, 500) 
        
    def announce_pcells(self, pc_ixys, dur=None):
        """ Anounce what we're up against
        200 ms first cell, 100 ms second cell 50 ms
        :pc_ixys: list of possible cells ixy
        :dur: max duration of beep default: 200 
        """
        if self.has_sinewave:
            self.sinewave_beep.announce_pcells(pc_ixys, dur)
        else:
            if dur is None:
                dur = self.beep_dur
            for pc_ixy in pc_ixys:
                self.announce_pcell(pc_ixy, dur)
                dur //= 2
        
    def announce_next_pcells(self, pc_ixys, dur=None):
        """ Announce what we're up against
        200 ms first cell, 100 ms second cell 50 ms
        :pc_ixys: list of possible cells ixy
        :dur: max duration of beep default: 200 
        """
        if self.has_sinewave:
            self.sinewave_beep.announce_next_pcells(pc_ixys, dur)
        else:
            if dur is None:
                dur = self.beep_dur
            for pc_ixy in pc_ixys:
                self.announce_pcell(pc_ixy, dur)
                dur //= 2

    def announce_next_pcell(self, pc_ixy):
        """ Announce next cell
        :pc_ixy: (ix,iy) of next cell
        """
        if self.has_sinewave:
            self.sinewave_beep.announce_next_pcell(pc_ixy=pc_ixy)
        else:    
            self.announce_pcell(pc_ixy=pc_ixy,
                                 dur=int(self.beep_dur*.5))
        
    def announce_pcell(self, pc_ixy, dur=None, dly=None):
        """ Announce cell
        :pc_ixy: ix,iy indexes
        :dur: duration default: pass on down
        :dly: delay before the sound default: pass on down
        """
        if self.silence():
            return
        
        if self.has_sinewave:
            self.sinewave_beep.announce_pcell(pc_ixy, dur=dur, dly=dly)
        else:
            if dur is None:
                dur = self.beep_dur
            cell = self.awin.get_cell_at_ixy(pc_ixy)
            if cell is not None:
                color = cell._color
                if color not in self.color_freqs:
                    freq = self.color_freqs["OTHER"]
                    dur = self.other_dur
                    SlTrace.lg(f"color:{color} freq:{freq} dur:{dur}",
                                "pos_tracking")
                else:
                    freq = self.color_freqs[color]
                winsound.Beep(freq, dur)
                SlTrace.lg(f"in cell: winsound.Beep({freq},{dur})"
                           f" cell:{cell} at {pc_ixy}", "pos_tracking")
            elif self.out_of_bounds_check(pc_ixy):
                SlTrace.lg(f"announce_pcell({pc_ixy}) out of bounds", "bounds")
            else:
                winsound.Beep(frequency=self.blank_freq, duration=self.blank_dur)
                SlTrace.lg(f"blank: winsound.Beep({self.blank_freq},{self.blank_dur})"
                           f" {cell} at {pc_ixy}", "pos_tracking")
            
    def out_of_bounds_check(self, pc_ixy):
        """ Check for out of bounds / illegal location
            and announce
        :pc_ixy: (ix,iy) location
        :returns: True if out of bounds / illegal
        """
        if pc_ixy is None:
            self.announce_can_not_do()
            return True
        
        ix, iy = pc_ixy
        if ix < self.awin.get_ix_min():
            self.announce_can_not_do()
            return True
        if ix > self.awin.get_ix_max():
            self.announce_can_not_do()
            return True
        if iy < self.awin.get_iy_min():
            self.announce_can_not_do()
            return True
        if iy > self.awin.get_iy_max():
            self.announce_can_not_do()
            return True

        return False


# adw_perimeter.py    04Apr2023  crs, Author
"""
Create / Manipulate figure perimiters
"""
import re 

from select_trace import SlTrace
from braille_cell import BrailleCell 



class SurroundingSquare:
    """ A potential surrounding square
    """
    def __init__(self, ssqs, ixy, prev_square=None, next_square=None):
        self.ssqs = ssqs
        self.adwp = ssqs.adwp
        self.ixy = ixy
                                        # Traversal links
        self.prev_square = prev_square  # (SurroundingSquare)
        self.next_square = next_square  # (SurroundingSquare)

    def __str__(self):
        st = f"SurroundingSquare:{self.ixy}"
        return st 
    
    def expand_if_possible(self):
        """ move cell, expanding surrounding cells, if possible
        :returns: True if expansion made, else False
        """
        expansion_conn_spaces = self.get_expansion_conn_spaces()
        if len(expansion_conn_spaces) > 0:
            expansion_conn_space = expansion_conn_spaces[0]   # take highest prioity
            self.move_to_expansion(expansion_conn_space)
            return True 
        
        return False
        
     
    def get_expansion_conn_spaces(self):
        """ Get possible expansion spaces
        return expand_spaces
        :returns: list of (connectors,space_ixy)
        """
        return self.adwp.get_expansion_conn_spaces(ixy=self.ixy)

    def is_neighbor(self, sp_ixy):
        """ Check if square is our neighbor
            Could be out of bounds
        :sp_ixy: candidate's location tuple
        :returns: True iff a neighbor
        """
        n_ixys = self.get_n_ixys(ixy=sp_ixy)
        return self.ixy in n_ixys

    def move_to_expansion(self, conn_space):
        """ Adjust the self.ssqs.surounding squares to convert ixy
        square to surrounding and, possibly remove current square,
        placing it in the outside region
        FOR NOW we won't reduce, but rather let the reduction pass
        take care of that.
        WE, FOR NOW, assume one and only one connectors is
        flat with the expansion space
        :conn_space: (conns, space_ixy) tuple:
                        connected surrounding squares
                        candidate for expansion ixy
        """
        conns, space_ixy = conn_space
        sp_ix, sp_iy = space_ixy
        if len(conns) == 2:
            """ Inside space becomes new surrounding square
                adjacent to flat_conn
                NOTE we do not do any minimizing surrounding
                at this time.
            """
            
            prev_conn = conns[0]
            new_ssquare = SurroundingSquare(ssqs=self.ssqs, ixy=space_ixy)
            self.ssqs.insert_square(square=new_ssquare,
                                           after_ixy=prev_conn.ixy)
        elif len(conns) == 3:       # In the middle?
            prev_trav = conns[0]
            new_surr = SurroundingSquare(ssqs=self.ssqs, ixy=space_ixy)
            self.ssqs.insert_square(square=new_surr, after_ixy=prev_trav.ixy)
            """ NOTE: we do not do any minimizing surrounding at this time
            """

        
    """
    links to ssqs - SurroundingSquares
    """
       
    def get_square(self, ixy):
        """ Get (SurroundingSquare) at ixy
        :ixy: (ix,iy) location
        :returns: SurroundingSquare
        """
        return self.ssqs.squares[ixy]

        
    """
    links to adwp
    """

    def get_n_ixys(self,ixy=None):
        """ Get neighboring index (ixy) pairs
        :ixy: middle index pair
        :returns: list of (ixy) surrounding ixy
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.get_n_ixys(ixy=ixy)
    
    def get_inside_space_neighbors(self, ixy=None):
        """ Get neighbor space squares
        :ixy: square ix,iy tuple
                default: our ixy
        :returns: list of neighboring spaces
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.get_inside_space_neighbors(ixy=ixy)
    
    def is_inside(self, ixy=None):
        """ Check if square(ixy) is in outside region
        :ixy: square ix,iy tuple
                default: our ixy
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.is_inside(ixy=ixy)

    def is_outside(self, ixy=None):
        """ Check if square(ixy) is in outside region
        :ixy: square ix,iy tuple
                default: our ixy
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.is_outside(ixy=ixy)

    def is_surrounding(self, ixy=None):
        """ Check if square(ixy) is in surrounding region
        :ixy: square ix,iy tuple
                default: our square
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.is_surrounding(ixy=ixy)
    
    def reduce_extra(self, ixy=None):
        """ Reduce cell if not necessary for surrounding,
        by making it an outside
        :ixy: ix,iy tuple
                default: self
        :returns: True iff reduced
        """
        if ixy is None:
            ixy = self.ixy
        return self.adwp.reduce_extra(ixy=ixy)
        
    def remove_surrounding_square(self, ixy=None, square=None):
        """ remove ixy or move square
            ONLY ixy for new or square for moved
        :ixy: (ixy) ixy of square
        :square: (SurroundingSquare) moved square
        :returns: removed square
        """
        return self.adwp.remove_surrounding_square(ixy=ixy, square=square)

class SurroundingSquares:
    """ Holder of list of squares empty or not, e.g. outside, surrounding
    This can be used to manipulate surrounding regions and perimeters of
    a figure.
    Primary data components are:
        1. dictionary, by ixy, of squares
        2. traversal list made up of:
            a. starting_ixy
            b. doublely linked list contained in the
                SurroundingSquare entry
                    before_ixy - refering  to the previous SurroundingSquare
                    after_ixy - revering to the following SurroundingSquare

    """
    def __init__(self, adwp, cells=None):
        """ Setup
        :adwp: AdwPerimeter instance
        :cells: dictionary of cells or list with .ix,.iy
            default: use self.cells
            if list (with elements having .ix,.iy) create
            dictionary by (ix,iy)
        """
        self.adwp = adwp
        if cells is None:
            cells = self.cells
        if isinstance(cells, list):
            cdict = {}
            for cell in cells:
                ixy = (cell.ix,cell.iy)
                cdict[ixy] = cell
            cells = cdict
        self.cells = cells
        self.squares = {}               # Dictionary by ix,iy
                                        # Links for traversal
        self.first_square = None        # First (SurroundSquare)
        
    def add_square(self, ixy=None, square=None):
        """ Add/Create ixy or move square
            ONLY ixy for new or square for moved
        :ixy: (ixy) ixy of new square
        :square: (SurroundingSquare) moved square
        """
        if ixy is None and square is None:
            raise Exception("One of ixy or square is Required") 
        if ixy is not None and square is not None:
            raise Exception("Including BOTH ixy and square")
        
        if ixy is not None: 
            ssq = SurroundingSquare(ssqs=self, ixy=ixy)
        elif square is not None:
            ssq = square
        self.insert_square(square=ssq)
       
    def insert_square(self, square, after_ixy=None):
        """ Insert square
        :square: (SurroundingSquare) to insert
        :afer_ixy:  tuple, after which we insert
                    default: insert at end of list
        """
        
        self.squares[square.ixy] = square
        if self.first_square is None:
            self.first_square = square  # Just store it
            return
        
        if after_ixy is None:
            after_square = self.first_square
            next_square = after_square
            while next_square is not None:
                after_square = next_square
                next_square = after_square.next_square
        else:
            after_square = self.get_square(ixy=after_ixy)
        if after_square is None:
            self.first_square = square
            prev_square = next_square = None
        else:
            prev_square = after_square
            next_square = after_square.next_square
            square.prev_square = prev_square
            square.next_square = next_square            
        if prev_square is not None:
            prev_square.next_square = square
        if next_square is not None:
            next_square.prev_square = square
        
    def remove_square(self, ixy=None, square=None):
        """ remove ixy or move square
            ONLY ixy for new or square for moved
            NOTE we relink previous neighbors
        :ixy: (ixy) ixy of square
        :square: (SurroundingSquare) moved square
        :returns: removed square
        """
        if ixy is None and square is None:
            raise Exception("One of ixy or square is Required") 
        if ixy is not None and square is not None:
            raise Exception("Including BOTH ixy and square")
        
        if ixy is not None: 
            ssq = self.squares[ixy]
        elif square is not None:
            ixy = square.ixy
            ssq = square            # return given square
            
        del self.squares[ixy]
        # Update linking references
        prev_square = ssq.prev_square
        next_square = ssq.next_square
        if prev_square is not None:
            prev_square.next_square = next_square
        if next_square is not None:
            next_square.prev_square = prev_square            
        return ssq
        
    def get_square(self, ixy):
        """ Get (SurroundingSquare) at ixy
        :ixy: (ix,iy) location
        :returns: SurroundingSquare if present else None
        """
        if ixy in self.squares:
            return self.squares[ixy]
        
        return None

    def get_squares_list(self):
        """ list of SurroundingSquare in traversal order
        :return: list of (SurroundingSquare)
        """
        squares_list = []
        ssq = self.first_square
        while ssq is not None:
            squares_list.append(ssq)
            ssq = ssq.next_square
            
        return squares_list
        
    def get_squares_as_cells(self, color=None):
        """ Return cell dictionary by ix,iy
        :color: cell color default: keep color of cell, else "x" if emptyl
        :returns: cell (BrailleCell) dictionary
        """
        cells = {}
        for sq_ixy in self.squares:
            if sq_ixy in self.cells:
                cell = self.cells[sq_ixy]
                if color is not None:
                    cell.color = color
            else:
                ix,iy = sq_ixy
                if color is None:
                    color = "x"
                cell = BrailleCell(ix=ix, iy=iy, color=color)
            cells[sq_ixy] = cell
        return cells

    def is_part (self, ixy):
        """ Check if square(ixy) is part of this region
        :ixy: square ix,iy tuple
        :returns: True if ixy is part of this region
        """
        is_part = ixy in self.squares
        return is_part

    
class AdwPerimeter:
    
    def __init__(self, adw, cells=None):
        """ Setup for perimeter calculation 
        :adw: Acces to AudioDrawWindow
        :cells: list/dictionary of BrailleCell by (ixy)
                default: adw cells (adw.get_cells()
        """
        self.adw = adw
        if cells is None:
            cells = self.get_cells()
        else:
            if isinstance(cells, list):
                cdict = {}
                for cell in cells:
                    cdict[(cell.ix,cell.iy)] = cell
                cells = cdict
        self.cells = cells
        
    def get_perimeter(self):
        """ Get perimeter list of cells
        Starts with:
            1. a rectangular region, labeled outside, of empty squares 2 squares
            outside the bounding box 2. a rectangular region, labeled
            surrounding, of empty squares 1 square outside the bounding box 3. a
            rectangular region, laveled inside, of all the squares within the
            bounding box
            
        The surrounding region squares are repeatedly traversed, resetting
        adjacent inside empty squares to surrounding squares and surrounding
        squares to outside squares so as to closely surround all remaining non-
        space inside squares as surrounding squares.
            
        When no more inside empty squares can be converted to surrounding,
        the perimeter consists of the non-empty inside squares adjacent to
        the surrounding squares.
             
        :returns: list of cells(BrailleCell) which is the perimeter of the figure             
        """
        surrounding,outside = self.get_surroundings()
        self.surrounding = surrounding
        self.outside = outside
        loop_count = 0
        while True:
            loop_count += 1
            self.print_text(title=f"get_perimetr loop {loop_count}")
            # Minimize surrounding squares to one square thick
            minimize_loop_count = 0
            while True:
                minimize_loop_count += 1
                self.print_text(
                    title=f"minimize loop {minimize_loop_count}")
                minimize_count = 0
                ssquare_list = self.surrounding.get_squares_list()
                for ssquare in ssquare_list:
                    if ssquare.reduce_extra():
                        minimize_count += 1
                if minimize_count == 0:
                    break           # No improvement - quit minimization
                
            ssquare_list = self.surrounding.get_squares_list()
            new_count = 0
            for ssquare in ssquare_list:
                SlTrace.lg(f"sq: {ssquare.ixy}", "track_perimeter")
                if ssquare.expand_if_possible():
                    new_count += 1
            if new_count == 0:
                break           # No more expansion

        ssquare_list = self.surrounding.get_squares_list()
        if SlTrace.trace("ssquare_list"):
            st = ""
            for ssquare in ssquare_list:
                st += f" {ssquare}"
            SlTrace.lg("ssquare list - closest surrounding squares")
            SlTrace.lg(st)
            
        perimeter_cell_list = self.surrounding_to_perimeter(ssquare_list)
        if SlTrace.trace("ssquare_list"):
            st = ""
            for cell in perimeter_cell_list:
                st += f" {cell.pi_number}: {cell}"
            SlTrace.lg("perimeter cell list")
            SlTrace.lg(st)
        
        return perimeter_cell_list

    def surrounding_to_perimeter(self, ssquares):
        """ Given surrounding, a list of SurroundingSquare adjacent to
         get adjacent cells
        :ssquares: list of SurroundingSquares
        :returns: list of cells in traversal order
                Sets perimeter cells
                         pi_number=position number
                         pi_type = "p"
        """
        cells = self.get_cells()
        cells_used = {}
        perimeter_poss = []     # Possible squares for each possition
        for ssquare in ssquares:
            SlTrace.lg(f"ssquare: {ssquare.ixy}", "ssquares_list")
            poss_for_sq = {}
            st = ""
            for cell in self.get_cell_neighbors(ssquare.ixy):
                ixy = (cell.ix,cell.iy)
                poss_for_sq[ixy] = cell
                st += f" {cell}"
            SlTrace.lg(f"    poss: {st}", "ssquares_list")
            perimeter_poss.append(poss_for_sq)
        cell_list = []
        
        pi_number = 0           # incremented to traversal position
        for sq_p in perimeter_poss:
            for cell_ixy in sq_p:
                if cell_ixy not in cells_used:
                    cell = cells[cell_ixy]
                    pi_number += 1
                    cell.pi_number = pi_number
                    cell.pi_type = "p"
                    cell_list.append(cell)
                    cells_used[cell_ixy] = cell_ixy
                    break
        return cell_list        
        
    def get_surroundings(self):
        """ create one square thick surrounding square list plus an adjacent
        a surrounding set of outside squares to aid "directing" movement away
        form the outside squares toward the inside squares

        It is assumed that the figure of non-space squares is at least
        two cells within the display limits.  If not the outer two squares
        will be treated as empty.

        :returns: tuple (
                surrounding empty squaress (SurroundingSquares),
                
                outside_squares (SurroundingSquares)
                    of squares encompasing  the surrounding
                    empty squares returned
                )
                ALSO sets figure boundary, plus edge:
                    self.ix_min, self.iy.min
                    self.ix_max, self.iy_max
        """
        # edge of 2:
        #            outer: encompassing outside squares
        #            inner: starting list of surrounding squares
         
        ix_min,iy_min, ix_max,iy_max = self.bounding_box_ci(
                        cells=self.cells, add_edge=2)
        self.ix_min = ix_min
        self.iy_min = iy_min
        self.ix_max = ix_max
        self.iy_max = iy_max
        surrounding = SurroundingSquares(adwp=self, cells=self.cells)
        outside = SurroundingSquares(adwp=self, cells=self.cells)
        # top edge outside
        iy = iy_min
        for ix in range(ix_min, ix_max+1):
            ixy = (ix,iy)
            outside.add_square(ixy)
        # top edge surrounding
        iy = iy_min+1
        for ix in range(ix_min+1, ix_max):
            ixy = (ix,iy)
            surrounding.add_square(ixy)
            
        # right edge outside
        ix = ix_max
        for iy in range(iy_min, iy_max+1):
            ixy = (ix,iy) 
            outside.add_square(ixy)
        # right edge surrounding
        ix = ix_max-1
        for iy in range(iy_min+1, iy_max):
            ixy = (ix,iy) 
            surrounding.add_square(ixy)
                            
        # bottom edge outside
        iy = iy_max
        for ix in range(ix_max, ix_min-1, -1):
            ixy = (ix,iy)
            outside.add_square(ixy)
        # bottom edge surrounding
        iy = iy_max-1
        for ix in range(ix_max-1, ix_min, -1):
            ixy = (ix,iy)
            surrounding.add_square(ixy)
            
        # left edge outside
        ix = ix_min
        for iy in range(iy_max, iy_min-1, -1):
            ixy = (ix,iy) 
            outside.add_square(ixy)
        # left edge surrounding
        ix = ix_min+1
        for iy in range(iy_max-1, iy_min, -1):
            ixy = (ix,iy) 
            surrounding.add_square(ixy)
        
        return (surrounding, outside)
           
    def get_start_cell(self):
        """ Get starting cell
        :returns: BrailleCell at leftest cell at lowest row 
        """
        ix_min,iy_min, ix_max,iy_max = self.bounding_box_ci(cells=self.cells)
        for ix in reversed(range(ix_min, ix_max+1)):
            ixy = (ix-1,iy_min) 
            if self.is_space(ixy):
                break       # ixy is left most  
        
        return self.get_cell_at_ixy(cell_ixy=ixy)

    def get_cell(self, ixy, cells=None):
        """ Get cell(BrailleCell) at (ixy) if one else None
        :ixy: ix,iy pair
        :cells: cells to search default: self.cells
        :returns: BrailleCell iff one else None
        """
        if cells is None:
            cells = self.cells
        return self.get_cell_at_ixy(cell_ixy=ixy, cells=cells)


    def is_outside(self, ixy):
        """ Check if square(ixy) is in outside region
        :ixy: square ix,iy tuple
        """
        return self.outside.is_part(ixy=ixy)

    def is_inside(self, ixy):
        """ Check if square(ixy) is inside figure
        that is within figure boundaries and not is_outside()
        and not is_surrounding()
        :ixy: square ix,iy tuple
        :returns: True iff inside else False
        """
        if not self.is_in_figure_boundary(ixy=ixy):
            return False

        if self.is_surrounding(ixy=ixy):
            return False 
        
        if self.is_outside(ixy=ixy):
            return False 
        
        return True

    def is_in_figure_boundary(self, ixy):
        """ Check if within bounding rectangle + edge
        of most recent get_surrounding() call
        :ixy: ix,iy tuple
        :returns: True if within bounding rectangle
        """
        ix,iy = ixy
        if ix < self.ix_min:
            return False 
        
        if ix > self.ix_max:
            return False 
        
        if iy < self.iy_min:
            return False 
        
        if iy > self.iy_max:
            return False 
        
        return True
        
    def is_surrounding(self, ixy):
        """ Check if square(ixy) is surrounding region
        :ixy: square ix,iy tuple
        """
        return self.surrounding.is_part(ixy=ixy)
        
    def is_touching_inside(self, ixy):
        """ Check if square(ixy) is next to inside
        :ixy: square ix,iy tuple
        """
        for n_ixy in self.get_n_ixys(ixy=ixy):
            if self.is_inside(n_ixy):
                return True
        
    def is_touching_outside(self, ixy):
        """ Check if square(ixy) is next to outside
        :ixy: square ix,iy tuple
        """
        for n_ixy in self.get_n_ixys(ixy=ixy):
            if self.is_outside(n_ixy):
                return True
                

    def add_edge_cell(self, ixy):
        """ Add edge cell to dictionary
        :ixy: ixy coordinate
        """
        cell = self.get_cell_at_ixy(cell_ixy=ixy)
        self.edge_cells[ixy] = cell
        
        
        
    def get_cell_neighbors(self, ixy):
        """ get neighbors that are non-empty cells
        :ixy: (ix,iy) cell index
        :returns: list of BrailleCell
        """
        ngh_ixys = self.get_n_ixys(ixy=ixy)
        ngh_cells = []
        for ixy in ngh_ixys:
            cell = self.get_cell_at_ixy(ixy)
            if cell is not None:
                 ngh_cells.append(cell)
        return ngh_cells

     
    def get_expansion_conn_spaces(self, ixy):
        """ Get possible expansion spaces for surrounding square
        To be a candidate surrounding space, we must have two concecutive
        surrounding spaces between which this candidate may be inserted
        to maintain a connected string of surrounding squares
        
        :ixy: our location tuple
        :returns: list of (surrounding connectors, inside space)
        Must be:
            1. adjacent (touching us)
            2. inside
            3. Have our adjacent surrounding neighbors as neighbors.
        That is this expansion must not break the chain
        of surrounding squares
        [s1]   [i]    2 conn neighbors =>  [s1]  [s1-s2] new surrounding
        *s2*]  [ ]                         *s2*  [ ]    
        [s3]   [ ]                         [s3]  [ ]

        [s1]  [ ]                          [s1]  [ ]
        *s2*  [i]    3 conn neighbors =>   *o*   [s1-s3] s2 at new location    
        [s3]  [ ]                          [s3]  [ ]

        [s1]  [ ]                          [s1]  [ ]
        *s2*  [ ]                          *s2*  [ ]    
        [s3]  [i]    2 conn neighbors =>   [s3]  [s2-s3] new surrounding

        :returns: list of possible expansion spaces [(cons, sp_ixy),...]
        decreasing probability
        NO prioritization yet TBD
        """
        
        inside_neighbors = self.get_inside_space_neighbors(ixy=ixy)
        if len(inside_neighbors) == 0:
            return []
        
        expand_spaces = []
        our_connectors = self.get_surrounding_connectors(ixy=ixy)
        st = [str(s) for s in our_connectors]
        SlTrace.lg(f"our_connectors:{st}")
        for sp_ixy in inside_neighbors:
            if True:
                SlTrace.lg(f"expansion_space:{sp_ixy}")
                self.print_text(marker_ixy=sp_ixy)
            conn_neighbors = []
            for conn in our_connectors:
                if conn.is_neighbor(sp_ixy):
                    conn_neighbors.append(conn)
            if len(conn_neighbors) > 1:
                expand_spaces.append((conn_neighbors,sp_ixy)) 
        return expand_spaces

    def get_surrounding_connectors(self, ixy):
        """ Get connecting (part of the chain)of which we are apart
        including our self (1,2, or 3)
        :ixy: ix,iy tuple  default: our ixy
        :returns: list of squares of connectors
        """
        conns = []
        square = self.surrounding.get_square(ixy=ixy)
        neighbors = [square.prev_square, square, square.next_square]
        for neighbor in neighbors:
            if neighbor is not None:
                if neighbor.is_surrounding():
                    conns.append(neighbor)
                
        return conns

    def insert_surrounding_square(self, square, after_ixy=None):
        """ Insert square into surrounding
        :square: (SurroundingSquare) to insert
        :afer_ixy:  tuple, after which we insert
                    default: insert at end of list
        """
        self.surrounding.insert_square(square=square,
                                        after_ixy=after_ixy)

        
    def remove_surrounding_square(self, ixy=None, square=None):
        """ remove ixy or move square
            ONLY ixy for new or square for moved
        :ixy: (ixy) ixy of square
        :square: (SurroundingSquare) moved square
        :returns: removed square
        """
        return self.surrounding.remove_square(ixy=ixy, square=square)

        
    def get_inside_space_neighbors(self, ixy):
        """ get neighbors that are non-empty cells
        :ixy: (ix,iy) cell index
        :returns: list of (ix,iy)  tuples which are empty spaces
        """
        ngh_ixys = self.get_n_ixys(ixy=ixy)
        ngh_spaces = []
        for ngh_ixy in ngh_ixys:
            if self.is_inside(ngh_ixy):
                cell = self.get_cell_at_ixy(ngh_ixy)
                if cell is None:
                    ngh_spaces.append(ngh_ixy)
        return ngh_spaces

    def get_n_ixys(self,ixy):
        """ Get neighboring index (ixy) pairs
        :ixy: middle index pair
        :returns: list of (ixy) surrounding ixy
        """
        nixys = []
        ix_m,iy_m = ixy
        for ix in range(-1,2):
            for iy in range(-1,2):
                if ix != 0 or iy != 0:
                    ix_s = ix_m + ix
                    iy_s = iy_m + iy
                    nixys.append((ix_s, iy_s))
        return nixys

        
    def get_text(self, cells=None, shift_to_edge=None, blank_char= ",",
                 marker_ixy=None, marker=None, marker_space=".",
                 cell_nch=4):
        """ get text picture for braille display from regions.
        This function was adapted from get_text member in BrailleCellText.
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :marker_ixy: ixy tuple to mark default: No marking
        :marker: Type marking
                "uppercase" - uppercase the expected character
                ONE char - e.g. X - use this character
                default: "uppercase"
        :marker_space: Marker for space default: "."
        :cell_nch: Number of characters per cell default: 1
        """
        if cells is None:
            cells = self.get_cells()
        if shift_to_edge is None:
            shift_to_edge = True
        self.blank_char = blank_char
        left_edge, top_edge, right_edge, bottom_edge =self.bounding_box_ci(
            cells=cells, add_edge=2)
        
        if not shift_to_edge:
            left_edge = 0
            top_edge = 0

        braille_text = ""
        for iy in range(top_edge, bottom_edge+1):
            line = ""
            for ix in range(left_edge, right_edge+1):
                cell_ixy = (ix,iy)
                if cell_ixy in cells:
                    cell = self.cells[cell_ixy]
                    color = cell.color_string()
                    mark = color[0]
                elif self.is_outside(cell_ixy):
                    mark = "-"      # Outside
                    if  marker_ixy == cell_ixy:
                        mark = marker_space
                elif self.is_surrounding(cell_ixy):
                    mark = "+"
                elif self.is_inside(cell_ixy):
                    mark = ":"
                else:
                    mark = '@'      # Other
                if cell_ixy == marker_ixy:
                    if mark == " ":
                        mark = marker_space
                    elif mark in BrailleCell.color_for_character:
                         mark.upper()
                    elif mark in "-+:":
                        mark = f"[{mark}]"
                    else:
                        mark = mark[0]
                if len(mark) == 1 and cell_nch > 1 and mark[0] != "[]":
                    mark *= cell_nch
                while len(mark) < cell_nch:
                    mark += "_"

                line += mark
            line = line.rstrip()
            if self.blank_char != " ":
                line = line.replace(" ", self.blank_char)
            ###print(f"{iy:2}", end=":")
            if not re.match(r"^\s*$", line):
                braille_text += line + "\n"
        return braille_text
    
    def print_text(self, title=None,
                   shift_to_edge=None, blank_char= ",",
                 marker_ixy=None, marker=None, marker_space="."):
        """ Print out (to log) text rendition
        :title: optional title default: no title
        :shift_to_edge: shift figure towards edge to ease finding figure
                        default: self.shift_to_edge
        :blank_char: convert leading blanks to this
                    default: ","
        :marker_ixy: ixy tuple to mark default: No marking
        :marker: Type marking
                "uppercase" - uppercase the expected character
                ONE char - e.g. X - use this character
                default: "uppercase"
        :marker_space: Marker for space default: "."
        """
        if title is None:
            title = ""
        SlTrace.lg(f"\n{title}")
        text = self.get_text(shift_to_edge=shift_to_edge,
                              blank_char= blank_char,
                              marker_ixy=marker_ixy, marker=marker,
                              marker_space=marker_space)
        SlTrace.lg(f"\n{text}")
        pass
    
    
    def reduce_extra(self, ixy):
        """ Reduce cell if not necessary for surrounding,
        by making it an outside
        :ixy: ix,iy tuple
                default: REQUIRED
        :returns: True iff reduced
        """
        if not self.is_surrounding(ixy=ixy):
            return False 
        
        if not self.is_touching_outside(ixy=ixy):
            return False
        
        if self.is_touching_inside(ixy=ixy):
            return False
        
        ssquare = self.remove_surrounding_square(ixy=ixy)
        self.outside.add_square(square=ssquare)
        return True
    
    def set_cell_ip_type(self, ixy=None, square=None, type=None):
        """ Set/clear perimeter type
        One and only one of cell_ixy, sauare is allowed
        :cell_ixy: cell ix,iy tuple
        :square: SurroundingSquare 
        :type: perimeter type i - inside, o - outside, p - perimeter
        """
        assert not (cell_ixy is None and square is None), "One must be here"
        assert not (cell_ixy is not None and square is not None), "Only one"
        if cell_ixy is None:
            cell_ixy = square.ixy
        cell = self.get_cell(cell_ixy)
        cell.ip_type = type
        
    """
    ############################################################
                       Links to adw
    ############################################################
    """

    def annotate_cell(self, cell_ixy=None, color=None,
                      outline="blue", outline_width=2,
                      text=None):
        """ Annotate cell to highlight it
        Possibly for perimeter viewing
        :cell_xy: ix,iy tuple default: current location
        :color: rectangle color default: no fill
        :outline: add outline color
                    default: no special outline
        :outline_width: outline width
                    default: 2
        :text: added text default: no text added
        """
        self.adw.annotate_cell(cell_ixy=cell_ixy,
                               color=color, outline=outline,
                               outline_width=outline_width,
                               text=text)
    
    def bounding_box_ci(self, cells=None, add_edge=None):
        """ cell indexes which bound the list of cells
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :add_edge: number of cells to add/subtract (if possible)
                     to enlarge/shrink box
                    default: no change
        :returns: 
                    None,None,None,None if no figure
                    upper left ix,iy  lower right ix,iy
                    that is ix_min,iy_min, ix_max,iy_max
        """
        return self.adw.bounding_box_ci(cells=cells, add_edge=add_edge)


    def find_edges(self):
        """Find  top, left, bottom, right non-blank edges
        so we can shift picture to left,top for easier
        recognition
        :returns: left_edge, top_edge, right_edge, bottom_edge
                    Also sets self.left_edge,...
        """
        return self.adw.find_edges()
    
            
    def get_cell_at_ixy(self, cell_ixy, cells=None):
        """ Get cell at (ix,iy), if one
        :cells: dictionary of cells by (ix,iy)
                default: self.get_cells()
        :cell_ixy: (ix,iy)
        :returns: BrailleCell if one, else None
        """
        return self.adw.get_cell_at_ixy(cell_ixy=cell_ixy, cells=cells)
    
            
    def get_ix_min(self):
        """ get minimum ix on grid
        :returns: min ix
        """
        return self.adw.get_ix_min()

    def get_ix_max(self):
        """ get maximum ix on grid
        :returns: max ix
        """
        return self.adw.get_ix_max()

    def get_iy_min(self):
        """ get minimum iy on grid
        :returns: min iy
        """
        return self.adw.get_iy_min()

    def get_iy_max(self):
        """ get maximum ix on grid
        :returns: max iy
        """
        return self.adw.get_iy_max()

    def get_cells(self):
        """ Get cell dictionary (by (ix,iy)
        """
        return self.adw.get_cells()
    
    def is_space(self, ixy=None):
        """ Are we at a space
        :ixy: cell ix,iy indexes 
        :returns: True if a space (not a cell) 
        """
        return self.adw.is_space(ixy=ixy)
    
    
if __name__ == "__main__":
    from braille_cell_text import BrailleCellText
    from audio_draw_window import AudioDrawWindow
    
    ts1_str = """
    
     rrr 
    ooooo
     ggg
     
"""
     
    ts2_str = """
    
    
    
                        r 
                       rrr
                       rrr 
                       rrr
                ooooooooryyyyyyyyy
                ooooooooryyyyyyyyy
                       ggg 
                       ggg
                       ggg
                        g
                        
                         
    """
    
    ts3_str = """
     
,,,,,,,,,,,iii
,,,,,,,,,,iiiii
,,,,,,,,,,iiiii,,,,,,vvv
,,,,,,,,,,iiiii,,,,,vvvvv
,,,,,,,,,,,,ii,,,,,,vvvvv
,,,bb,,,,,,,,i,,,,,,vvvvv
,,bbbbb,,,,,,i,,,,,vv
,,bbbbb,,,,,,i,,,,vv
,,bbbbbbb,,,,ii,,vv
,,,,,,,,bbbb,,i,vv,,,,,,,,rr
,,,,,,,,,,bbbbivv,,,,,,,,rrrr
,,,,,,,,,,,,,bvvrrrrrrrrrrrrr
,,,,,,,,,,ggggyoo,,,,,,,,rrrr
,,,,,,,,gggg,,y,oo,,,,,,,,rr
,,ggggggg,,,,yy,,oo
,,ggggg,,,,,,y,,,,oo
,,ggggg,,,,,,y,,,,,oo
,,,gg,,,,,,,,y,,,,,,ooooo
,,,,,,,,,,,,yy,,,,,,ooooo
,,,,,,,,,,yyyyy,,,,,ooooo
,,,,,,,,,,yyyyy,,,,,,ooo
,,,,,,,,,,yyyyy
,,,,,,,,,,,yyy
"""

    ts4_str = """
,,ggrrrrrro
,,ggrrrrrro
,,ggrrrrrro
,,gg,,,,,,o
,,gg,,,,,,o
,,gg,,,,,,o
,,gg,,,,,,o
,,ggyyyyyyy
,,ggyyyyyyy
"""
    adw = None
    test_it_num = 0
    def test_it(figure_str, desc=None):
        """ Test figure producing perimeter
        :figure_str: figure text string
        :desc: test decription default: generated
        """
        global adw
        global test_it_num
        cell_nch = 4
        test_it_num += 1
        if desc is None:
            desc = ""
        title = f"{desc} test: {test_it_num}"
        adw = AudioDrawWindow(title=title)
        SlTrace.lg(f"\n{title}")
        bct = BrailleCellText(text=figure_str, cell_nch=4)
        cells = bct.get_cells()
        adw.draw_cells(cells=cells)
        SlTrace.lg("\nOriginal Figure")
        bct.print_text()
        figure_cells = bct.get_cells()    # Avoiding origin change
        ssq = AdwPerimeter(adw)
        perimeter_cells = ssq.get_perimeter()
        perim_num = 0
        for cell in perimeter_cells:
            perim_num += 1
            ssq.annotate_cell(cell_ixy=(cell.ix,cell.iy))
            
        
        for pcell in perimeter_cells:
            pcell_ixy = (pcell.ix,pcell.iy)
            pcolor = pcell.color_string().upper()
            pcell.color = "violet"
            figure_cells[pcell_ixy] = pcell
        pp_bct = BrailleCellText(cells=figure_cells, cell_nch=cell_nch)
        pp_bct_text = pp_bct.get_text()
        SlTrace.lg(f"With perimeter accented")
        SlTrace.lg(pp_bct_text)
        
    SlTrace.clearFlags()
    #test_it(figure_str=ts1_str, desc="simplest")    
    test_it(figure_str=ts2_str, desc="colored cross")    
    test_it(figure_str=ts3_str, desc="spokes")    
    test_it(figure_str=ts4_str, desc="square colors")    

    adw.mainloop()          # To keep AudioDrawWindow responsive



#adw_scanner.py    09Mar2023  crs, Author
"""
Scanning support
The idea is to rapidly scan the grid giving audio tonal feedback as to the 
contents
"""
import math
import time
import copy
import cProfile, pstats, io         # profiling support

from sinewave_numpy import SineWaveNumPy

from select_trace import SlTrace
from sinewave_beep import SineWaveBeep
from adw_perimeter import AdwPerimeter

"""
scan path item to facilitate fast display, sound and operation
"""
class ScanPathItem:
    
    def __init__(self, ix, iy, cell=None,
                 sinewave_numpy=None,      # if not scan_use_tone
                 pitch=None,
                 dur=None,
                 dly=None,
                 sinewave_left=None,
                 sinewave_right=None):
        """ Setup scan item
        :ix: ix index
        :iy: iy index
        :cell: BrailleCell if one else None
        """
        self.ix = ix
        self.iy = iy
        self.cell = cell
        self.sinewave_numpy = sinewave_numpy  # Setup after init
        self.pitch = pitch          # Setup if scan_use_tone
        self.dur = dur              # Setup if scan_use_tone
        self.dly = dly
        self.sinewave_left = sinewave_left      # ??? OBSOLETE
        self.sinewave_right = sinewave_right    # ??? OBSOLETE

    def __str__(self):
        st = f"ScanPathItem: {self.ix},{self.iy}"
        st += f" {self.cell}" if self.cell is not None else " space"
        return st
    
class AdwScanner:

    def __init__(self, fte, cell_time=.1, space_time=.1,
                 skip_space=True, skip_run=True, scan_len=10,
                 skip_run_max = 10,        # Maximum to run skip at a time
                 skip_space_max = 3,             # Maximum to space skip at a time
                 sample_rate = 44100,
                 space_pitch=None,
                 no_item_wait=False, scan_use_tone=True,
                 scan_coverage="perimeter",
                 add_tone_preamble=False,
                 combine_wave=False,
                 n_combine_wave=2):
        """ Setup scanning
        :fte: front end instance
        :cell_time: time (seconds) to beep cell
        :space_time: time (seconds) to beep space
        :skip_space: True - skip space default: True
        :skip_space_max: maximum number of spaces to skip per time
        :skip_run: True - skip run of equals default: True
        :skip_run_max: maximum number run per time
        :scan_len: number of items to add to scan list
        :sample_rate: default tone sample rate
        :space_pitch: scanning space pitch default: "SCAN_BLANK"
        :no_item_wait: True - don't wait for items to complete
        :scan_use_tone: True use play_tone for scanning
                        same as cursor movements
                        default: True
        :scan_coverage: cell coverage strategy
                            "area" - cover rectangular region
                            "perimeter" - cover perimeter of figure
                        default: perimeter
        :add_tone_preamble: Add tone to sync beginning of scan wave
                        default: No tone preamble added
        :combine_wave: waveforms are combined to speed presentation
                        default: False
        :n_combine_wave: number of sections combined wave 
                        default: 4
        """
        #space_time = 1      # TFD
        #cell_time = 2      # TFD
        self.fte = fte
        self.adw = fte.adw
        self.speaker_control = self.adw.speaker_control
        self.mw = self.adw.mw
        self.canvas = self.adw.canvas
        self.cell_time = cell_time
        self.space_time = space_time
        self.sample_rate = sample_rate
        self._scan_item = None          # Scanning image
        self._display_item = None       # Currently displayed item, if any
        self._scan_item_tag = None             # Scanning item tag
        self._report_item = None        # Latest scanning report item
        self._scanning = False          # Currently scanning
        self.set_skip_space(skip_space)
        self.set_skip_space_max(skip_space)
        self.set_skip_run(skip_run)
        self.set_skip_run_max(skip_run_max)
        self.set_scan_len(scan_len)
        self.skip_color = None          # color of run if any
        self._display_item = None
        self.profile_running = False    # Setup disabled profiling
        self.set_no_item_wait(no_item_wait)
        self._in_display_item = False   # Avoid recursion
        self._in_update = 0
        self.sound_queue_back_log = 1  # Target to keep
        self.scan_loop_proc_time_ms = 200   # queue cking loop time
        self.scan_use_tone = scan_use_tone
        self.add_tone_preamble = add_tone_preamble
        self.set_combine_wave(combine_wave)
        self.n_combine_wave = n_combine_wave
        if space_pitch is None:
            self.space_pitch = SineWaveBeep.color2pitch("SCAN_SPACE")
            self.scan_loop_checking = None      # set if after alive
        self.scan_coverage = scan_coverage

    def set_cell_time(self, time):
        """ Set cell tone duration hoped
        :time: cell time in seconds
        """
        self.cell_time = time

    def set_space_time(self, time):
        """ Set space tone duration hoped
        :time: cell time in seconds
        """
        self.space_time = time

    def set_space_pitch(self, pitch):
        """ Set space tone duration hoped
        :tone: space tone (pitch)
        """
        self.space_pitch = pitch

    def set_combine_wave(self, val=True):
        """ Set/clear combine_wave mode
        :val: set combine_wave default: True - do combine wave
        """
        self.combine_wave = val

    def set_scan_len(self, scan_len):
        """ Set number of items to current scan list
        :scan_len: number of items to add each trip
        """
        self.scan_len = scan_len

    def flip_skip_space(self):
        """ Flip skipping spaces
        """
        skip_space = self.get_skip_space()
        self.set_skip_space(not skip_space)


    def is_skip_space(self):
        return self.skip_space

    def get_skip_space(self):
        return self.skip_space

    def set_skip_space(self, val):
        self.skip_space = val

    def set_skip_space_max(self, val):
        self.skip_space_max = val


    def is_skip_run(self):
        return self.skip_run

    def set_skip_run(self, val):
        self.skip_run = val

    def set_skip_run_max(self, val):
        self.skip_run_max = val

    def flip_skip_run(self):
        """ Flip skipping runs
        """
        skip_run = self.is_skip_run()
        self.set_skip_run(not skip_run)

    def set_scanning(self, cells=None):
        """ Setup for scanning or any other cell based operation
        which requires cell placement informations
        e.g. volume vs position
        :cells: cells to scan
            default: self.get_cells()
        """
        if cells is None:
            cells  = self.get_cells()
        self.scan_cells = cells
        eye_sep_half = 15    # Half separation
        eye_ix = (self.get_ix_min() + self.get_ix_max())//2
        if eye_ix > self.get_ix_min() - eye_sep_half:
            eye_ix_l = eye_ix - eye_sep_half
        else:
            eye_ix_l = self.get_ix_min()
        if eye_ix < self.get_ix_max() - eye_sep_half:
            eye_ix_r = eye_ix + eye_sep_half
        else:
            eye_ix_r = self.get_ix_max()

        eye_iy_l = eye_iy_r = self.get_iy_max()
        self.eye_ixy_l = (eye_ix_l, eye_iy_l)
        self.eye_ixy_r = (eye_ix_r, eye_iy_r)
        fig_x_ul,fig_y_ul,fig_x_lr,fig_y_lr = self.bounding_box_ci()
        self.view_left = fig_x_ul if fig_x_ul <= eye_ix_l else eye_ix_l
        self.view_right  = fig_x_lr if fig_x_lr >= eye_ix_r else eye_ix_r

    def start_scanning(self, cells=None):
        """ Start scanning process
        eye(center) is bottom / middle
        """
        SlTrace.lg("start_scanning")
        self.set_scanning()
        SlTrace.lg("Calculating scan_items")
        ###self.forward_path = self.get_scan_path(cells=self.scan_cells)
        ###self.reverse_path = list(reversed(self.forward_path))
        self.setup_get_scan_path()
        self.more_to_get = True
        SlTrace.lg(f"start_scan ")
        self.start_scan()


    def setup_get_scan_path(self, cells=None):
        """ Setup getting path in parts
        :cells: cells to scan
        """
        self.scan_items = self.get_scan_items(cells=cells)
        self.scan_items_index = 0
        self.scan_items_end = len(self.scan_items)      # < end

    def get_scan_items(self, cells=None):
        """ Get ordered list of scanning locations, skipping based
        on settings self.skip_space, self.skip_run
        :cells: cells dictionary by ix,iy default: self.cells
        :returns: Rqw list of ScanPathItem in the order to scan
        """
        scan_items_raw = self.get_scan_items_raw(cells=cells)
        items_raw_left = scan_items_raw[:]    # Depleated as we go
        scan_items = []         # Populated with used scan items

        run_color = None                            # color in run
        while len(items_raw_left) > 0:              # Loop over raw items
            n_skip = 0                              # avoid too long a skip
            while len(items_raw_left) > 0:          # In skipping loop
                item = items_raw_left.pop(0)
                if self.skip_space and item.cell is None:
                    if n_skip == 0:
                        scan_items.append(item) # Include first space
                    else:
                        if len(items_raw_left) > 0:
                            next_item = items_raw_left[0]
                            if next_item.cell is not None:
                                #scan_items.append(item)     # Include last space
                                break                       # Quit space run
                    n_skip += 1
                    if n_skip > self.skip_space_max:
                        scan_items.append(item)     # Add in space in long 
                        break
                    continue

                elif self.skip_run and item.cell is not None:
                    if len(items_raw_left) == 0:
                        break               # Don't skip end of path
                    color = item.cell.color_string()
                    if run_color is None or run_color != color:
                        run_color = color
                        break                   # Starting run
                    next_item = items_raw_left[0]
                    if next_item.cell is None or next_item.cell.color_string() != color:
                        break       # Include end of run
                    if next_item.cell.iy != item.cell.iy:
                        break       # Include new row
                    n_skip += 1
                    if n_skip > self.skip_run_max:
                        break
                else:
                    break
                SlTrace.lg(f"skipping {item}", "scanning")
            SlTrace.lg(f"scanning {item}", "scanning")
            if item.cell is None:
                SlTrace.lg(f"None: scanning {item}")

            scan_items.append(item)
        if SlTrace.trace("scanning"):
            SlTrace.lg(f"get_scan_items returns:")
            for si in scan_items:
                SlTrace.lg(f"  {si}")
        return scan_items

    def get_scan_items_raw(self, cells=None):
        """ Get ordered list of scanning locations, no skipping
            Scanning order is bottom to top, with alternating
            left-to-right then right-to-left so as to make a contiguous
            path
        :cells: cells dictionary by ix,iy default: self.cells
        :returns: Rqw list of ScanPathItem in the order to scan
        """
        if cells is None:
            cells = self.get_cells()
        if len(cells) == 0:
            return []       # No cells to scann

        if self.scan_coverage == "perimeter":
            self.skip_run = False   # skip_run is too aggressive
            scan_items = self.get_perimeter_items(cells=cells)
            return scan_items

        ix_ul, iy_ul, ix_lr, iy_lr = self.bounding_box_ci(cells=cells)
        scan_items = []
        left_to_right = True    # Start going left to right
        for iy in range(iy_lr, iy_ul-1, -1):    # bottom to top
            if left_to_right:
                for ix in range(ix_ul, ix_lr+1):
                    cell = cells[(ix,iy)] if (ix,iy) in cells else None
                    scan_items.append(ScanPathItem(ix=ix, iy=iy, cell=cell))
            else:
                for ix in reversed(range(ix_ul, ix_lr+1)):
                    cell = cells[(ix,iy)] if (ix,iy) in cells else None
                    scan_items.append(ScanPathItem(ix=ix, iy=iy, cell=cell))
            left_to_right = not left_to_right
        return scan_items

    def get_perimeter_items(self, cells=None):
        """ get items on the perimeter
        :cells: cells on which perimeter is based
                default: self.get_cells()
        :returns: Raw list of ScanPathItem based on perimeter
        """
        if cells is None:
            cells = self.get_cells()
        adwp = AdwPerimeter(adw=self.adw, cells=cells)
        cell_list = adwp.get_perimeter()
        scan_items = []
        for cell in cell_list:
            ix,iy = (cell.ix,cell.iy)
            scan_items.append(ScanPathItem(ix=ix, iy=iy, cell=cell))
        return scan_items

    def get_more_scan_path(self, nitem=10, use_sinewave_numpy=False):
        """ Get ordered list of scanning locations
            Scanning order is bottom to top, with alternating
            left-to-right then right-to-left so as to make a contiguous
            path
        :cells: cells dictionary by ix,iy default: self.cells
        :nitem; number of items to return
        :use_sinewave_numpy: if present add sinewave_numpy entry to each item
        :returns: list of ScanPathItem in the order to scan at most
                    nitem per call
                    Last return len < nitem
        """
        """ Populate with sinewave appropriate with locations
        We assume we can hold enough pysinewave instances each set for stereo.
        """
        SlTrace.lg(f"Add {nitem} new items at {self.scan_items_index} with sound")
        if self.profile_running:
            self.pr = cProfile.Profile()
            self.pr.enable()
        begin_time = time.time()
        next_items = []
        vol_adj = self.get_vol_adj()
        SlTrace.lg(f"Scanning vol_adj: {vol_adj}")
        for _ in range(nitem):
            if self.scan_items_index >= self.scan_items_end:
                return next_items
            sp_ent = self.scan_items[self.scan_items_index]
            next_items.append(sp_ent)
            self.scan_items_index += 1
            ix,iy = sp_ent.ix,sp_ent.iy
            volume = self.get_vol(ix=ix, iy=iy)
            vol_left, vol_right = volume
            vol_adj = self.get_vol_adj()
            if abs(vol_adj) > 1.e-10:  # Only if non-trivial
                vol_left += vol_adj
                vol_right += vol_adj
            if sp_ent.cell is None:
                pitch = self.space_pitch
                item_time = self.space_time
            else:
                pitch = self.get_pitch(ix=ix, iy=iy)
                item_time = self.cell_time
            sp_ent.pitch = pitch
            sp_ent.dur = item_time
            sp_ent.vol = (vol_left, vol_right)
            if not self.scan_use_tone or use_sinewave_numpy:
                sinewave_numpy = SineWaveNumPy(pitch = pitch,
                                               duration=item_time,
                                               decibels_left=vol_left,
                                               decibels_right=vol_right)
                sp_ent.sinewave_numpy = sinewave_numpy
        end_time = time.time()
        dur_time = end_time-begin_time
        if self.profile_running:
            self.pr.disable()
            s = io.StringIO()
            sortby = 'cumulative'
            ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
            ps.print_stats()
            SlTrace.lg(s.getvalue())

        SlTrace.lg(f"{nitem} cells: time: {dur_time:.3f} seconds")
        return next_items

    def get_pitch(self, ix, iy):
        """ Get pitch for cell (color)
        :ix: cell ix
        :iy: cell iy
        """
        cell = self.get_cell_at_ixy((ix,iy))
        color = cell.color_string()
        pitch = self.color2pitch(color)
        return pitch

    def get_vol(self, ix, iy, eye_ixy_l=None, eye_ixy_r=None):
        """ Get tone volume for cell at ix,iy
        volume(left,right)
        Volume(in decibel): k1*(k2-distance from eye), k1=1, k2=0
        :ix: cell ix
        :iy: cell iy
        :eye_xy_l: left eye/ear at x,y default: self.eye_xy_l
        :eye_xy_r: right eye/ear at x,y  default: self.eye_xy_r
        return: volume(left,right) in decibel
        """
        if eye_ixy_l is None:
            eye_ixy_l = self.eye_ixy_l
        if eye_ixy_r is None:
            eye_ixy_r = self.eye_ixy_r
        eye_ix_l, eye_iy_l = eye_ixy_l
        eye_ix_r, eye_iy_r = eye_ixy_r
        eye_ix = (eye_ix_l+eye_ix_r)/2
        eye_iy = (eye_iy_l+eye_iy_r)/2
        dist = math.sqrt((ix-eye_ix)**2 + (iy-eye_iy)**2)   # average dist
        k1 = 15
        k1 = 2     # TFD - quieter
        k2 = 10
        kd = .5
        vol = k1*(k2-dist*kd)
        SILENT = -100
        width = abs(self.view_right - self.view_left)
        vol_l = (vol*abs(self.view_right-ix)/width
                 + SILENT*abs(self.view_left-ix)/width)
        vol_r = (vol*abs(self.view_left-ix)/width
                 + SILENT*abs(self.view_right-ix)/width)
        '''
        if vol_l > vol_r:
            vol_r = SILENT           # HACK to enphasize stereo
        else:
            vol_l = SILENT
        vol_l,vol_r = vol_r,vol_l   # HACK - got things reversed!
        '''
        SlTrace.lg(f"\nl-r: {vol_l-vol_r:.2f}", "sound_volume")
        SlTrace.lg(f"{ix},{iy} left vol:{vol_l:.2f}  dist:{dist:.2f} eye_ixy_l:{eye_ixy_l}",
                   "sound_volume")
        SlTrace.lg(f"{ix},{iy} right vol:{vol_r:.2f}  dist:{dist:.2f} eye_ixy_r:{eye_ixy_r}",
                   "sound_volume")
        return vol_l,vol_r


    def start_scan(self):
        """ Start actualscanning, which continues until stop_scan
        is called.
        Path self.forward_path is used, then alternating self.reverse_path,
        and forward_path
        """
        self._scanning = True
        self.first_scan = True      # Used if combine_wave
        self.using_forward_path = True
        self.forward_path = []
        self.reverse_path = []
        self.current_scan_path = []     # let scan_path_item setup
        self.scan_loop_checking = self.mw.after(0, self.scan_loop_proc)
        self.mw.after(0,self.scan_path_item)

    def scan_loop_proc(self):
        """ Keep speaker_control moderately full of scan path items
        """
        if self._scanning and not self.combine_wave:
            #while (self.get_sound_queue_size()
            #        < self.sound_queue_back_log):
            #    self.scan_path_item()
            if self.get_sound_queue_size() < 1:
                self.scan_path_item()
            self.scan_loop_checking = self.mw.after(self.scan_loop_proc_time_ms,
                                                    self.scan_loop_proc)

    def stop_scan(self):
        """ Stop current scan
        """
        if not self._scanning:
            return      # Not scanning - no disrupting 

        self._scanning = False
        if self.scan_loop_checking is not None:
            self.mw.after_cancel(self.scan_loop_checking)
            self.scan_loop_checking = None
        self.remove_display_item()
        self.remove_report_item()
        self.speaker_control.stop_scan()


    def scan_path_item(self):
        """Process next entry in self.current_path, removing item from
        list.  If at end of path, add next path, switching between
        forward_path and reverse_path
        """
        if not self._scanning:
            return

        if self.combine_wave:
            self.do_combine_wave()
            return

        self.more_to_get = True if (
                self.scan_items_index < self.scan_items_end) else False
        item = None             # Set iff found
        if len(self.current_scan_path) == 0:
            if self.more_to_get:
                new_items = self.get_more_scan_path(nitem=self.scan_len)
                self.forward_path.extend(new_items)
                self.reverse_path = list(reversed(self.forward_path))
                self.current_scan_path = new_items   # do new ones
            else:
                if self.scan_coverage == "perimeter":
                    self.current_scan_path = self.forward_path[:]   # used up
                else:
                    if self.using_forward_path:
                        self.current_scan_path = self.forward_path[:]   # used up
                        self.using_forward_path = False
                    else:
                        self.current_scan_path = self.reverse_path[:]   # used up
                        self.using_forward_path = True

        if len(self.current_scan_path) > 0:
            item = self.current_scan_path.pop(0)
            SlTrace.lg(f"\nitem: {item}", "scanning")
            self._display_item = item
            self.display_item(item)
            if self.scan_use_tone:
                self.speaker_control.play_tone(
                    pitch=item.pitch, dur=item.dur,
                    volume=item.vol, dly=item.dly)
                self.update()
                self.speaker_control.wait_while_busy()
                self.update()

            else:
                ###if self.no_item_wait and not self.more_to_get:
                if self.no_item_wait:
                    self.report_item(item, no_wait = True)
                    self.mw.after(int(self.cell_time*1000),
                                  self.scan_path_item_complete)
                else:
                    self.report_item(item, after=self.scan_path_item_complete)


    def scan_path_item_complete(self):
        """ Complete current item where necessary
        """
        self.report_item_complete()
        if not self.no_item_wait:
            if self._scanning:
                self.mw.after(0, self.scan_path_item)

    def do_combine_wave(self):
        """ Do combined wave scan
            creating wave(s) before first scan
        """
        if not self._scanning:
            return

        if self.first_scan:
            self.first_scan = False
            self.wave_index = -1        # Bumped before do_wave
            if self.scan_loop_checking is not None:
                self.mw.after_cancel(self.scan_loop_checking)
                self.scan_loop_checking = None
            self.forward_path = self.get_more_scan_path(nitem=
                                                        len(self.scan_items),
                                                        use_sinewave_numpy=True)
            self.forward_paths = self.divide_path(self.forward_path,
                                                  n_section=self.n_combine_wave)
            if self.add_tone_preamble:
                self.forward_path[0] = self.add_preamble(self.forward_path[0])
            self.forward_waves = []
            for path in self.forward_paths:
                wave = self.make_combine_wave(path)
                self.forward_waves.append(wave)
        self.wave_index += 1
        if self.wave_index >= self.n_combine_wave:
            self.wave_index = 0     # Restart around
        self.do_wave(self.forward_waves[self.wave_index],
                     index=self.wave_index)
        '''
        else:
            self.do_wave(self.reverse_wave)
        self.using_forward_path = not self.using_forward_path
        '''

    def divide_path(self, path, n_section):
        """ Divide scan path into almost even sections
        :path: scan path
        :n_section: number of sections
        :returns: list of list of ScanPathItems
        """
        path_sects = []         # path sections
        items_left = path[:]    # Depopulate using all but no more
        n_per_sect = len(path)//n_section
        if n_per_sect*n_section < len(path):
            n_per_sect += 1     # Use them up
        for _ in range(n_section):
            path_sect = []
            for _ in range(n_per_sect):
                if len(items_left) > 0:
                    path_sect.append(items_left.pop(0))
            path_sects.append(path_sect)
        SlTrace.lg(f"divide_path: {len(path)} items in {n_section} sections")
        for i in range(n_section):
            item = path_sects[i][0]
            SlTrace.lg(f"{i}: row: {item.iy+1} col: {item.ix+1}")

        return path_sects


    def add_preamble(self, items, ptype="BEGINNING"):
        """ Add preamble to item list to aid in identification
        :items: item list of (ScanPathItem)
        :ptype: preamble type default: "BEGINNING"
        """
        items_new = items[:]
        # beginning
        if len(items) == 0:
            item_model = ScanPathItem(ix=0, iy=0, sinewave_numpy=None)
            SlTrace.lg("Empty list")
        else:
            item_model = copy.copy(items[0])
        ix_start,iy_start = item_model.ix,item_model.iy
        sinewave_numpy = item_model.sinewave_numpy
        if sinewave_numpy is None:
            sample_rate = self.sample_rate
        else:
            sample_rate =  sinewave_numpy.sample_rate
        nitem = 5
        cpsp = SineWaveBeep.cpsp        # Pitch spacing
        pitch_start = SineWaveBeep.color2pitch("SPACE") + 15*cpsp
        preamble_items = []
        vol_adj = self.get_vol_adj()
        SlTrace.lg(f"Scanning vol_adj: {vol_adj}")
        for i in range(nitem):
            ix = ix_start
            iy = iy_start-i
            volume = self.get_vol(ix=ix, iy=iy)
            vol_left, vol_right = volume
            pitch = pitch_start
            duration = self.cell_time*2
            if abs(vol_adj) > 1.e-10:  # Only if non-trivial
                vol_left += vol_adj
                vol_right += vol_adj
            sinewave_numpy = SineWaveNumPy(pitch = pitch,
                                           duration=duration,
                                           decibels_left=vol_left,
                                           decibels_right=vol_right,
                                           sample_rate=sample_rate)
            preamble_item = ScanPathItem(ix=ix, iy=iy,
                                         sinewave_numpy=sinewave_numpy)
            preamble_items.append(preamble_item)
        items_new[:0] = preamble_items[:]
        return items_new

    def do_wave(self, scan_wave,  index):
        """ Process sound wave, returning when completed
        :index:  index into self.forward_path for identification
        """
        if not self._scanning:
            return

        current_path = self.forward_paths[index]
        if len(current_path) > 0:
            current_item = current_path[0]
            current_ix,current_iy = current_item.ix,current_item.iy
            self.speak_text(msg=f"r{current_iy+1} c{current_ix+1}",
                            dup_stdout=False, rate=350, volume=1)
        self.play_waveform(sinewave_numpy=scan_wave, calculate_dur=True,
                           after=self.do_wave_end)
        #self.speaker_control.wait_while_busy()

    def do_wave_end(self):
        if not self._scanning:
            return

        self.mw.after(0, self.do_combine_wave)

    def make_combine_wave(self, item_path):
        """ Create a SinewaveNumpy.sinewave_numpy ndarray of the items components
        :itempath: list of ScanPathItem s with  sinewave_numpy   field set 
                        Note that get_scan_items_raw needs self.
        :returns: SpeakerWaveform of combined item waveforms
        """
        SlTrace.lg(f"make_combine_wave: {len(item_path)} items")
        swnps = [sp_ent.sinewave_numpy for sp_ent in item_path]
        cw_swnp = SineWaveNumPy.concatinate(swnps)
        return cw_swnp

    def color2pitch(self, color):
        return SineWaveBeep.color2pitch(color)


    def display_item(self, item):
        """ Update current scanner position display
        :item: current scan path item
        """
        if self._in_display_item:
            return

        if self._display_item is not None:
            self.remove_display_item()
        canvas = self.canvas
        ixy = item.ix, item.iy
        x_left,y_top, x_right,y_bottom = self.get_win_ullr_at_ixy_canvas(ixy)
        self._scan_item_tag = canvas.create_rectangle(x_left,y_top,
                                                      x_right,y_bottom,
                                                      outline="black", width=2)
        self.update()
        self._display_item = item
        self._in_display_item = False

    def remove_display_item(self):
        """ remove current display item, if any
        """
        if  self._display_item is not None:
            canvas = self.canvas
            if self._scan_item_tag is not None:
                self.canvas.delete(self._scan_item_tag)
                self._scan_item_tag = None
        self._display_item = None

    def remove_report_item(self):
        """ Remove any noise for reported item
        """
        if self._report_item is not None:
            self._report_item.sinewave_numpy.stop()
            self._report_item = None        # TBD - turn tone off

    def report_item(self, item, after=None, no_wait=False):
        """ Report scanned item
        start tone, to be stopped via report_item_complete
        :item: item being reported via tone
        :after: Function to call (win.after) after report completes
                default: no call
        :no_wait: True - no waiting for completion
                default: False - wait for completion
        """
        if no_wait:
            #self.remove_display_item()  # Remove previous item, if any
            if item is not None and item.cell is not None:
                sinewave_numpy = item.sinewave_numpy
                if sinewave_numpy is not None:
                    self._report_item = item
                    self.play_waveform(sinewave_numpy=sinewave_numpy)
            self.mw.after(int (self.cell_time*1000), self.scan_path_item_complete)
        else:
            sinewave_numpy = item.sinewave_numpy
            if sinewave_numpy is not None:
                self._report_item = item
                self.play_waveform(sinewave_numpy=sinewave_numpy,
                                   after=after)
            else:
                self.mw.after(0, after) # Just go on


    def report_item_complete(self):
        """ Report scanned item
        start tone, to be stopped via report_item_complete
        """
        self.remove_report_item()


    def stop_scanning(self):
        SlTrace.lg("stop_scanning")
        self.stop_scan()

    def set_no_item_wait(self, val=True):
        """ set/clear no_item_wait mode
        :val: True set value True default: True
        """
        self.no_item_wait = val


    def set_profile_running(self, val=True):
        """ Set/clear profiler running
        :val: True set profiling default: True
        """
        self.profile_running = val

    """
    ############################################################
                       Links to fte
    ############################################################
    """

    """ None so far """

    """
    ############################################################
                       Links to adw
    ############################################################
    """

    def bounding_box_ci(self, cells=None):
        """ cell indexes which bound the list of cells
        :cells: list of cells, (with cell.ix,cell.iy) or (ix,iy) tuples
                default: list of all cells in figure
        :returns: 
                    None,None,None,None if no figure
                    upper left ix,iy  lower right ix,iy
        """
        return self.adw.bounding_box_ci(cells=cells)

    def get_cell_at_ixy(self, cell_ixy):
        """ Get cell at (ix,iy), if one
        :cell_ixy: (ix,iy)
        :returns: BrailleCell if one, else None
        """
        return self.adw.get_cell_at_ixy(cell_ixy=cell_ixy)

    def get_cells(self):
        """ Get cell dictionary (by (ix,iy)
        """
        return self.adw.cells

    def get_speaker_control(self):
        """ Get speech control
        """
        return self.speaker_control

    def get_xy_canvas(self, xy=None):
        """ Get current xy pair on canvas (0-max)
        :xy: xy tuple default: current xy
        :returns: (x,y) tuple
        """
        return self.adw.get_xy_canvas(xy=xy)

    def set_grid_path(self):
        self.adw.set_grid_path()

    def get_grid_path(self):
        return self.adw.get_grid_path()

    def get_ix_min(self):
        """ get minimum ix on grid
        :returns: min ix
        """
        return self.adw.get_ix_min()

    def get_ix_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.adw.get_ix_max()

    def get_iy_min(self):
        """ get minimum iy on grid
        :returns: min iy
        """
        return self.adw.get_iy_min(())

    def get_iy_max(self):
        """ get maximum ix on grid
        :returns: min ix
        """
        return self.adw.get_iy_max()

    def get_win_ullr_at_ixy(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom)
        """
        return self.adw.get_win_ullr_at_ixy(ixy)


    def get_win_ullr_at_ixy_canvas(self, ixy):
        """ Get window rectangle for cell at ixy
        :ixy: cell index tupple (ix,iy)
        :returns: (x_left,y_top, x_right,y_bottom) canvas coords
        """
        return self.adw.get_win_ullr_at_ixy_canvas(ixy)

    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9            
        """
        self.adw.speak_text(msg=msg, msg_type=msg_type,
                            dup_stdout=dup_stdout,
                            rate=rate, volume=volume)

    def update(self):
        """ Update display
        """
        if self._in_update>3:
            return
        self._in_update += 1
        self.adw.update()
        self._in_update -= 1

    def mainloop(self):
        self.mw.mainloop()

    """
    ############################################################
                       Links to speaker_control
    ############################################################
    """

    def clear(self):
        """ Clear pending output
        """
        self.speaker_control.clear()

    def force_clear(self):
        """ Clear all pending and reset
        """
        self.speaker_control.force_clear()

    def get_vol_adj(self):
        """ Get current volume adjustment
        :returns: current vol_adjustment in db
        """
        return self.speaker_control.get_vol_adj()

    def get_sound_queue_size(self):
        return self.speaker_control.get_sound_queue_size()

    def play_waveform(self, sinewave_numpy,
                      calculate_dur=False,
                      after=None):
        """ play waveform
        :sinewave_numpy: waveform (SinewaveNumPy stereo_waveform)
        :after: function to call after play completes
                default: no call
        """
        swnp = sinewave_numpy
        self.speaker_control.play_waveform(ndarr=swnp.wf_ndarr,
                                           dur=swnp.duration,
                                           dly=swnp.delay,
                                           calculate_dur=calculate_dur,
                                           after=after)

    